# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from lxml import etree

from cinder.api import xmlutil
from cinder import test


class SelectorTest(test.TestCase):
    obj_for_test = {'test': {'name': 'test',
                             'values': [1, 2, 3],
                             'attrs': {'foo': 1,
                                       'bar': 2,
                                       'baz': 3, }, }, }

    def test_empty_selector(self):
        sel = xmlutil.Selector()
        self.assertEqual(len(sel.chain), 0)
        self.assertEqual(sel(self.obj_for_test), self.obj_for_test)

    def test_dict_selector(self):
        sel = xmlutil.Selector('test')
        self.assertEqual(len(sel.chain), 1)
        self.assertEqual(sel.chain[0], 'test')
        self.assertEqual(sel(self.obj_for_test),
                         self.obj_for_test['test'])

    def test_datum_selector(self):
        sel = xmlutil.Selector('test', 'name')
        self.assertEqual(len(sel.chain), 2)
        self.assertEqual(sel.chain[0], 'test')
        self.assertEqual(sel.chain[1], 'name')
        self.assertEqual(sel(self.obj_for_test), 'test')

    def test_list_selector(self):
        sel = xmlutil.Selector('test', 'values', 0)
        self.assertEqual(len(sel.chain), 3)
        self.assertEqual(sel.chain[0], 'test')
        self.assertEqual(sel.chain[1], 'values')
        self.assertEqual(sel.chain[2], 0)
        self.assertEqual(sel(self.obj_for_test), 1)

    def test_items_selector(self):
        sel = xmlutil.Selector('test', 'attrs', xmlutil.get_items)
        self.assertEqual(len(sel.chain), 3)
        self.assertEqual(sel.chain[2], xmlutil.get_items)
        for key, val in sel(self.obj_for_test):
            self.assertEqual(self.obj_for_test['test']['attrs'][key], val)

    def test_missing_key_selector(self):
        sel = xmlutil.Selector('test2', 'attrs')
        self.assertIsNone(sel(self.obj_for_test))
        self.assertRaises(KeyError, sel, self.obj_for_test, True)

    def test_constant_selector(self):
        sel = xmlutil.ConstantSelector('Foobar')
        self.assertEqual(sel.value, 'Foobar')
        self.assertEqual(sel(self.obj_for_test), 'Foobar')


class TemplateElementTest(test.TestCase):
    def test_element_initial_attributes(self):
        # Create a template element with some attributes
        elem = xmlutil.TemplateElement('test', attrib=dict(a=1, b=2, c=3),
                                       c=4, d=5, e=6)

        # Verify all the attributes are as expected
        expected = dict(a=1, b=2, c=4, d=5, e=6)
        for k, v in expected.items():
            self.assertEqual(elem.attrib[k].chain[0], v)

    def test_element_get_attributes(self):
        expected = dict(a=1, b=2, c=3)

        # Create a template element with some attributes
        elem = xmlutil.TemplateElement('test', attrib=expected)

        # Verify that get() retrieves the attributes
        for k, v in expected.items():
            self.assertEqual(elem.get(k).chain[0], v)

    def test_element_set_attributes(self):
        attrs = dict(a=None, b='foo', c=xmlutil.Selector('foo', 'bar'))

        # Create a bare template element with no attributes
        elem = xmlutil.TemplateElement('test')

        # Set the attribute values
        for k, v in attrs.items():
            elem.set(k, v)

        # Now verify what got set
        self.assertEqual(len(elem.attrib['a'].chain), 1)
        self.assertEqual(elem.attrib['a'].chain[0], 'a')
        self.assertEqual(len(elem.attrib['b'].chain), 1)
        self.assertEqual(elem.attrib['b'].chain[0], 'foo')
        self.assertEqual(elem.attrib['c'], attrs['c'])

    def test_element_attribute_keys(self):
        attrs = dict(a=1, b=2, c=3, d=4)
        expected = set(attrs.keys())

        # Create a template element with some attributes
        elem = xmlutil.TemplateElement('test', attrib=attrs)

        # Now verify keys
        self.assertEqual(set(elem.keys()), expected)

    def test_element_attribute_items(self):
        expected = dict(a=xmlutil.Selector(1),
                        b=xmlutil.Selector(2),
                        c=xmlutil.Selector(3))
        keys = set(expected.keys())

        # Create a template element with some attributes
        elem = xmlutil.TemplateElement('test', attrib=expected)

        # Now verify items
        for k, v in elem.items():
            self.assertEqual(expected[k], v)
            keys.remove(k)

        # Did we visit all keys?
        self.assertEqual(len(keys), 0)

    def test_element_selector_none(self):
        # Create a template element with no selector
        elem = xmlutil.TemplateElement('test')

        self.assertEqual(len(elem.selector.chain), 0)

    def test_element_selector_string(self):
        # Create a template element with a string selector
        elem = xmlutil.TemplateElement('test', selector='test')

        self.assertEqual(len(elem.selector.chain), 1)
        self.assertEqual(elem.selector.chain[0], 'test')

    def test_element_selector(self):
        sel = xmlutil.Selector('a', 'b')

        # Create a template element with an explicit selector
        elem = xmlutil.TemplateElement('test', selector=sel)

        self.assertEqual(elem.selector, sel)

    def test_element_subselector_none(self):
        # Create a template element with no subselector
        elem = xmlutil.TemplateElement('test')

        self.assertIsNone(elem.subselector)

    def test_element_subselector_string(self):
        # Create a template element with a string subselector
        elem = xmlutil.TemplateElement('test', subselector='test')

        self.assertEqual(len(elem.subselector.chain), 1)
        self.assertEqual(elem.subselector.chain[0], 'test')

    def test_element_subselector(self):
        sel = xmlutil.Selector('a', 'b')

        # Create a template element with an explicit subselector
        elem = xmlutil.TemplateElement('test', subselector=sel)

        self.assertEqual(elem.subselector, sel)

    def test_element_append_child(self):
        # Create an element
        elem = xmlutil.TemplateElement('test')

        # Make sure the element starts off empty
        self.assertEqual(len(elem), 0)

        # Create a child element
        child = xmlutil.TemplateElement('child')

        # Append the child to the parent
        elem.append(child)

        # Verify that the child was added
        self.assertEqual(len(elem), 1)
        self.assertEqual(elem[0], child)
        self.assertIn('child', elem)
        self.assertEqual(elem['child'], child)

        # Ensure that multiple children of the same name are rejected
        child2 = xmlutil.TemplateElement('child')
        self.assertRaises(KeyError, elem.append, child2)

    def test_element_extend_children(self):
        # Create an element
        elem = xmlutil.TemplateElement('test')

        # Make sure the element starts off empty
        self.assertEqual(len(elem), 0)

        # Create a few children
        children = [xmlutil.TemplateElement('child1'),
                    xmlutil.TemplateElement('child2'),
                    xmlutil.TemplateElement('child3'), ]

        # Extend the parent by those children
        elem.extend(children)

        # Verify that the children were added
        self.assertEqual(len(elem), 3)
        for idx in range(len(elem)):
            self.assertEqual(children[idx], elem[idx])
            self.assertIn(children[idx].tag, elem)
            self.assertEqual(elem[children[idx].tag], children[idx])

        # Ensure that multiple children of the same name are rejected
        children2 = [xmlutil.TemplateElement('child4'),
                     xmlutil.TemplateElement('child1'), ]
        self.assertRaises(KeyError, elem.extend, children2)

        # Also ensure that child4 was not added
        self.assertEqual(len(elem), 3)
        self.assertEqual(elem[-1].tag, 'child3')

    def test_element_insert_child(self):
        # Create an element
        elem = xmlutil.TemplateElement('test')

        # Make sure the element starts off empty
        self.assertEqual(len(elem), 0)

        # Create a few children
        children = [xmlutil.TemplateElement('child1'),
                    xmlutil.TemplateElement('child2'),
                    xmlutil.TemplateElement('child3'), ]

        # Extend the parent by those children
        elem.extend(children)

        # Create a child to insert
        child = xmlutil.TemplateElement('child4')

        # Insert it
        elem.insert(1, child)

        # Ensure the child was inserted in the right place
        self.assertEqual(len(elem), 4)
        children.insert(1, child)
        for idx in range(len(elem)):
            self.assertEqual(children[idx], elem[idx])
            self.assertIn(children[idx].tag, elem)
            self.assertEqual(elem[children[idx].tag], children[idx])

        # Ensure that multiple children of the same name are rejected
        child2 = xmlutil.TemplateElement('child2')
        self.assertRaises(KeyError, elem.insert, 2, child2)

    def test_element_remove_child(self):
        # Create an element
        elem = xmlutil.TemplateElement('test')

        # Make sure the element starts off empty
        self.assertEqual(len(elem), 0)

        # Create a few children
        children = [xmlutil.TemplateElement('child1'),
                    xmlutil.TemplateElement('child2'),
                    xmlutil.TemplateElement('child3'), ]

        # Extend the parent by those children
        elem.extend(children)

        # Create a test child to remove
        child = xmlutil.TemplateElement('child2')

        # Try to remove it
        self.assertRaises(ValueError, elem.remove, child)

        # Ensure that no child was removed
        self.assertEqual(len(elem), 3)

        # Now remove a legitimate child
        elem.remove(children[1])

        # Ensure that the child was removed
        self.assertEqual(len(elem), 2)
        self.assertEqual(elem[0], children[0])
        self.assertEqual(elem[1], children[2])
        self.assertNotIn('child2', elem)

        # Ensure the child cannot be retrieved by name
        def get_key(elem, key):
            return elem[key]
        self.assertRaises(KeyError, get_key, elem, 'child2')

    def test_element_text(self):
        # Create an element
        elem = xmlutil.TemplateElement('test')

        # Ensure that it has no text
        self.assertIsNone(elem.text)

        # Try setting it to a string and ensure it becomes a selector
        elem.text = 'test'
        self.assertEqual(hasattr(elem.text, 'chain'), True)
        self.assertEqual(len(elem.text.chain), 1)
        self.assertEqual(elem.text.chain[0], 'test')

        # Try resetting the text to None
        elem.text = None
        self.assertIsNone(elem.text)

        # Now make up a selector and try setting the text to that
        sel = xmlutil.Selector()
        elem.text = sel
        self.assertEqual(elem.text, sel)

        # Finally, try deleting the text and see what happens
        del elem.text
        self.assertIsNone(elem.text)

    def test_apply_attrs(self):
        # Create a template element
        attrs = dict(attr1=xmlutil.ConstantSelector(1),
                     attr2=xmlutil.ConstantSelector(2))
        tmpl_elem = xmlutil.TemplateElement('test', attrib=attrs)

        # Create an etree element
        elem = etree.Element('test')

        # Apply the template to the element
        tmpl_elem.apply(elem, None)

        # Now, verify the correct attributes were set
        for k, v in elem.items():
            self.assertEqual(str(attrs[k].value), v)

    def test_apply_text(self):
        # Create a template element
        tmpl_elem = xmlutil.TemplateElement('test')
        tmpl_elem.text = xmlutil.ConstantSelector(1)

        # Create an etree element
        elem = etree.Element('test')

        # Apply the template to the element
        tmpl_elem.apply(elem, None)

        # Now, verify the text was set
        self.assertEqual(str(tmpl_elem.text.value), elem.text)

    def test__render(self):
        attrs = dict(attr1=xmlutil.ConstantSelector(1),
                     attr2=xmlutil.ConstantSelector(2),
                     attr3=xmlutil.ConstantSelector(3))

        # Create a master template element
        master_elem = xmlutil.TemplateElement('test', attr1=attrs['attr1'])

        # Create a couple of slave template element
        slave_elems = [xmlutil.TemplateElement('test', attr2=attrs['attr2']),
                       xmlutil.TemplateElement('test', attr3=attrs['attr3']), ]

        # Try the render
        elem = master_elem._render(None, None, slave_elems, None)

        # Verify the particulars of the render
        self.assertEqual(elem.tag, 'test')
        self.assertEqual(len(elem.nsmap), 0)
        for k, v in elem.items():
            self.assertEqual(str(attrs[k].value), v)

        # Create a parent for the element to be rendered
        parent = etree.Element('parent')

        # Try the render again...
        elem = master_elem._render(parent, None, slave_elems, dict(a='foo'))

        # Verify the particulars of the render
        self.assertEqual(len(parent), 1)
        self.assertEqual(parent[0], elem)
        self.assertEqual(len(elem.nsmap), 1)
        self.assertEqual(elem.nsmap['a'], 'foo')

    def test_render(self):
        # Create a template element
        tmpl_elem = xmlutil.TemplateElement('test')
        tmpl_elem.text = xmlutil.Selector()

        # Create the object we're going to render
        obj = ['elem1', 'elem2', 'elem3', 'elem4']

        # Try a render with no object
        elems = tmpl_elem.render(None, None)
        self.assertEqual(len(elems), 0)

        # Try a render with one object
        elems = tmpl_elem.render(None, 'foo')
        self.assertEqual(len(elems), 1)
        self.assertEqual(elems[0][0].text, 'foo')
        self.assertEqual(elems[0][1], 'foo')

        # Now, try rendering an object with multiple entries
        parent = etree.Element('parent')
        elems = tmpl_elem.render(parent, obj)
        self.assertEqual(len(elems), 4)

        # Check the results
        for idx in range(len(obj)):
            self.assertEqual(elems[idx][0].text, obj[idx])
            self.assertEqual(elems[idx][1], obj[idx])

    def test_subelement(self):
        # Try the SubTemplateElement constructor
        parent = xmlutil.SubTemplateElement(None, 'parent')
        self.assertEqual(parent.tag, 'parent')
        self.assertEqual(len(parent), 0)

        # Now try it with a parent element
        child = xmlutil.SubTemplateElement(parent, 'child')
        self.assertEqual(child.tag, 'child')
        self.assertEqual(len(parent), 1)
        self.assertEqual(parent[0], child)

    def test_wrap(self):
        # These are strange methods, but they make things easier
        elem = xmlutil.TemplateElement('test')
        self.assertEqual(elem.unwrap(), elem)
        self.assertEqual(elem.wrap().root, elem)

    def test_dyntag(self):
        obj = ['a', 'b', 'c']

        # Create a template element with a dynamic tag
        tmpl_elem = xmlutil.TemplateElement(xmlutil.Selector())

        # Try the render
        parent = etree.Element('parent')
        elems = tmpl_elem.render(parent, obj)

        # Verify the particulars of the render
        self.assertEqual(len(elems), len(obj))
        for idx in range(len(obj)):
            self.assertEqual(elems[idx][0].tag, obj[idx])


class TemplateTest(test.TestCase):
    def test_wrap(self):
        # These are strange methods, but they make things easier
        elem = xmlutil.TemplateElement('test')
        tmpl = xmlutil.Template(elem)
        self.assertEqual(tmpl.unwrap(), elem)
        self.assertEqual(tmpl.wrap(), tmpl)

    def test__siblings(self):
        # Set up a basic template
        elem = xmlutil.TemplateElement('test')
        tmpl = xmlutil.Template(elem)

        # Check that we get the right siblings
        siblings = tmpl._siblings()
        self.assertEqual(len(siblings), 1)
        self.assertEqual(siblings[0], elem)

    def test__nsmap(self):
        # Set up a basic template
        elem = xmlutil.TemplateElement('test')
        tmpl = xmlutil.Template(elem, nsmap=dict(a="foo"))

        # Check out that we get the right namespace dictionary
        nsmap = tmpl._nsmap()
        self.assertNotEqual(id(nsmap), id(tmpl.nsmap))
        self.assertEqual(len(nsmap), 1)
        self.assertEqual(nsmap['a'], 'foo')

    def test_master_attach(self):
        # Set up a master template
        elem = xmlutil.TemplateElement('test')
        tmpl = xmlutil.MasterTemplate(elem, 1)

        # Make sure it has a root but no slaves
        self.assertEqual(tmpl.root, elem)
        self.assertEqual(len(tmpl.slaves), 0)

        # Try to attach an invalid slave
        bad_elem = xmlutil.TemplateElement('test2')
        self.assertRaises(ValueError, tmpl.attach, bad_elem)
        self.assertEqual(len(tmpl.slaves), 0)

        # Try to attach an invalid and a valid slave
        good_elem = xmlutil.TemplateElement('test')
        self.assertRaises(ValueError, tmpl.attach, good_elem, bad_elem)
        self.assertEqual(len(tmpl.slaves), 0)

        # Try to attach an inapplicable template
        class InapplicableTemplate(xmlutil.Template):
            def apply(self, master):
                return False
        inapp_tmpl = InapplicableTemplate(good_elem)
        tmpl.attach(inapp_tmpl)
        self.assertEqual(len(tmpl.slaves), 0)

        # Now try attaching an applicable template
        tmpl.attach(good_elem)
        self.assertEqual(len(tmpl.slaves), 1)
        self.assertEqual(tmpl.slaves[0].root, good_elem)

    def test_master_copy(self):
        # Construct a master template
        elem = xmlutil.TemplateElement('test')
        tmpl = xmlutil.MasterTemplate(elem, 1, nsmap=dict(a='foo'))

        # Give it a slave
        slave = xmlutil.TemplateElement('test')
        tmpl.attach(slave)

        # Construct a copy
        copy = tmpl.copy()

        # Check to see if we actually managed a copy
        self.assertNotEqual(tmpl, copy)
        self.assertEqual(tmpl.root, copy.root)
        self.assertEqual(tmpl.version, copy.version)
        self.assertEqual(id(tmpl.nsmap), id(copy.nsmap))
        self.assertNotEqual(id(tmpl.slaves), id(copy.slaves))
        self.assertEqual(len(tmpl.slaves), len(copy.slaves))
        self.assertEqual(tmpl.slaves[0], copy.slaves[0])

    def test_slave_apply(self):
        # Construct a master template
        elem = xmlutil.TemplateElement('test')
        master = xmlutil.MasterTemplate(elem, 3)

        # Construct a slave template with applicable minimum version
        slave = xmlutil.SlaveTemplate(elem, 2)
        self.assertEqual(slave.apply(master), True)

        # Construct a slave template with equal minimum version
        slave = xmlutil.SlaveTemplate(elem, 3)
        self.assertEqual(slave.apply(master), True)

        # Construct a slave template with inapplicable minimum version
        slave = xmlutil.SlaveTemplate(elem, 4)
        self.assertEqual(slave.apply(master), False)

        # Construct a slave template with applicable version range
        slave = xmlutil.SlaveTemplate(elem, 2, 4)
        self.assertEqual(slave.apply(master), True)

        # Construct a slave template with low version range
        slave = xmlutil.SlaveTemplate(elem, 1, 2)
        self.assertEqual(slave.apply(master), False)

        # Construct a slave template with high version range
        slave = xmlutil.SlaveTemplate(elem, 4, 5)
        self.assertEqual(slave.apply(master), False)

        # Construct a slave template with matching version range
        slave = xmlutil.SlaveTemplate(elem, 3, 3)
        self.assertEqual(slave.apply(master), True)

    def test__serialize(self):
        # Our test object to serialize
        obj = {'test': {'name': 'foobar',
                        'values': [1, 2, 3, 4],
                        'attrs': {'a': 1,
                                  'b': 2,
                                  'c': 3,
                                  'd': 4, },
                        'image': {'name': 'image_foobar', 'id': 42, }, }, }

        # Set up our master template
        root = xmlutil.TemplateElement('test', selector='test',
                                       name='name')
        value = xmlutil.SubTemplateElement(root, 'value', selector='values')
        value.text = xmlutil.Selector()
        attrs = xmlutil.SubTemplateElement(root, 'attrs', selector='attrs')
        xmlutil.SubTemplateElement(attrs, 'attr', selector=xmlutil.get_items,
                                   key=0, value=1)
        master = xmlutil.MasterTemplate(root, 1, nsmap=dict(f='foo'))

        # Set up our slave template
        root_slave = xmlutil.TemplateElement('test', selector='test')
        image = xmlutil.SubTemplateElement(root_slave, 'image',
                                           selector='image', id='id')
        image.text = xmlutil.Selector('name')
        slave = xmlutil.SlaveTemplate(root_slave, 1, nsmap=dict(b='bar'))

        # Attach the slave to the master...
        master.attach(slave)

        # Try serializing our object
        siblings = master._siblings()
        nsmap = master._nsmap()
        result = master._serialize(None, obj, siblings, nsmap)

        # Now we get to manually walk the element tree...
        self.assertEqual(result.tag, 'test')
        self.assertEqual(len(result.nsmap), 2)
        self.assertEqual(result.nsmap['f'], 'foo')
        self.assertEqual(result.nsmap['b'], 'bar')
        self.assertEqual(result.get('name'), obj['test']['name'])
        for idx, val in enumerate(obj['test']['values']):
            self.assertEqual(result[idx].tag, 'value')
            self.assertEqual(result[idx].text, str(val))
        idx += 1
        self.assertEqual(result[idx].tag, 'attrs')
        for attr in result[idx]:
            self.assertEqual(attr.tag, 'attr')
            self.assertEqual(attr.get('value'),
                             str(obj['test']['attrs'][attr.get('key')]))
        idx += 1
        self.assertEqual(result[idx].tag, 'image')
        self.assertEqual(result[idx].get('id'),
                         str(obj['test']['image']['id']))
        self.assertEqual(result[idx].text, obj['test']['image']['name'])

    def test_serialize_with_delimiter(self):
        # Our test object to serialize
        obj = {'test': {'scope0:key1': 'Value1',
                        'scope0:scope1:key2': 'Value2',
                        'scope0:scope1:scope2:key3': 'Value3'
                        }}

        # Set up our master template
        root = xmlutil.TemplateElement('test', selector='test')
        key1 = xmlutil.SubTemplateElement(root, 'scope0:key1',
                                          selector='scope0:key1')
        key1.text = xmlutil.Selector()
        key2 = xmlutil.SubTemplateElement(root, 'scope0:scope1:key2',
                                          selector='scope0:scope1:key2')
        key2.text = xmlutil.Selector()
        key3 = xmlutil.SubTemplateElement(root, 'scope0:scope1:scope2:key3',
                                          selector='scope0:scope1:scope2:key3')
        key3.text = xmlutil.Selector()
        serializer = xmlutil.MasterTemplate(root, 1)
        xml_list = []
        xml_list.append("<?xmlversion='1.0'encoding='UTF-8'?><test>")
        xml_list.append("<scope0><key1>Value1</key1><scope1>")
        xml_list.append("<key2>Value2</key2><scope2><key3>Value3</key3>")
        xml_list.append("</scope2></scope1></scope0></test>")
        expected_xml = ''.join(xml_list)
        result = serializer.serialize(obj)
        result = result.replace('\n', '').replace(' ', '')
        self.assertEqual(result, expected_xml)


class MasterTemplateBuilder(xmlutil.TemplateBuilder):
    def construct(self):
        elem = xmlutil.TemplateElement('test')
        return xmlutil.MasterTemplate(elem, 1)


class SlaveTemplateBuilder(xmlutil.TemplateBuilder):
    def construct(self):
        elem = xmlutil.TemplateElement('test')
        return xmlutil.SlaveTemplate(elem, 1)


class TemplateBuilderTest(test.TestCase):
    def test_master_template_builder(self):
        # Make sure the template hasn't been built yet
        self.assertIsNone(MasterTemplateBuilder._tmpl)

        # Now, construct the template
        tmpl1 = MasterTemplateBuilder()

        # Make sure that there is a template cached...
        self.assertIsNotNone(MasterTemplateBuilder._tmpl)

        # Make sure it wasn't what was returned...
        self.assertNotEqual(MasterTemplateBuilder._tmpl, tmpl1)

        # Make sure it doesn't get rebuilt
        cached = MasterTemplateBuilder._tmpl
        tmpl2 = MasterTemplateBuilder()
        self.assertEqual(MasterTemplateBuilder._tmpl, cached)

        # Make sure we're always getting fresh copies
        self.assertNotEqual(tmpl1, tmpl2)

        # Make sure we can override the copying behavior
        tmpl3 = MasterTemplateBuilder(False)
        self.assertEqual(MasterTemplateBuilder._tmpl, tmpl3)

    def test_slave_template_builder(self):
        # Make sure the template hasn't been built yet
        self.assertIsNone(SlaveTemplateBuilder._tmpl)

        # Now, construct the template
        tmpl1 = SlaveTemplateBuilder()

        # Make sure there is a template cached...
        self.assertIsNotNone(SlaveTemplateBuilder._tmpl)

        # Make sure it was what was returned...
        self.assertEqual(SlaveTemplateBuilder._tmpl, tmpl1)

        # Make sure it doesn't get rebuilt
        tmpl2 = SlaveTemplateBuilder()
        self.assertEqual(SlaveTemplateBuilder._tmpl, tmpl1)

        # Make sure we're always getting the cached copy
        self.assertEqual(tmpl1, tmpl2)


class MiscellaneousXMLUtilTests(test.TestCase):
    def test_make_flat_dict(self):
        expected_xml = ("<?xml version='1.0' encoding='UTF-8'?>\n"
                        '<wrapper><a>foo</a><b>bar</b></wrapper>')
        root = xmlutil.make_flat_dict('wrapper')
        tmpl = xmlutil.MasterTemplate(root, 1)
        result = tmpl.serialize(dict(wrapper=dict(a='foo', b='bar')))
        self.assertEqual(result, expected_xml)



#    (c) Copyright 2014 Brocade Communications Systems Inc.
#    All Rights Reserved.
#
#    Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""Unit tests for FC Zone Manager."""

import mock

from cinder import exception
from cinder import test
from cinder.volume import configuration as conf
from cinder.zonemanager.drivers.fc_zone_driver import FCZoneDriver
from cinder.zonemanager.fc_zone_manager import ZoneManager
from mock import Mock

fabric_name = 'BRCD_FAB_3'
init_target_map = {'10008c7cff523b01': ['20240002ac000a50']}
fabric_map = {'BRCD_FAB_3': ['20240002ac000a50']}
target_list = ['20240002ac000a50']


class TestFCZoneManager(ZoneManager, test.TestCase):

    def setUp(self):
        super(TestFCZoneManager, self).setUp()
        self.configuration = conf.Configuration(None)
        self.configuration.set_default('fc_fabric_names', fabric_name)
        self.driver = Mock(FCZoneDriver)

    def __init__(self, *args, **kwargs):
        test.TestCase.__init__(self, *args, **kwargs)

    def test_add_connection(self):
        with mock.patch.object(self.driver, 'add_connection')\
                as add_connection_mock:
            self.driver.get_san_context.return_value = fabric_map
            self.add_connection(init_target_map)
            self.driver.get_san_context.assert_called_once(target_list)
            add_connection_mock.assert_called_once_with(fabric_name,
                                                        init_target_map)

    def test_add_connection_error(self):
        with mock.patch.object(self.driver, 'add_connection')\
                as add_connection_mock:
            add_connection_mock.side_effect = exception.FCZoneDriverException
            self.assertRaises(exception.ZoneManagerException,
                              self.add_connection, init_target_map)

    def test_delete_connection(self):
        with mock.patch.object(self.driver, 'delete_connection')\
                as delete_connection_mock:
            self.driver.get_san_context.return_value = fabric_map
            self.delete_connection(init_target_map)
            self.driver.get_san_context.assert_called_once_with(target_list)
            delete_connection_mock.assert_called_once_with(fabric_name,
                                                           init_target_map)

    def test_delete_connection_error(self):
        with mock.patch.object(self.driver, 'delete_connection')\
                as del_connection_mock:
            del_connection_mock.side_effect = exception.FCZoneDriverException
            self.assertRaises(exception.ZoneManagerException,
                              self.delete_connection, init_target_map)



# Copyright 2014 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Utilities for linking request ID's across service calls.
"""

import logging

from openstack.common.gettextutils import _  # noqa


LOG = logging.getLogger(__name__)


def link_request_ids(context, source_id, target_id=None, stage=None,
                     target_name=None, notifier=None):
    """Links the Request ID from the Source service to
       the Request ID returned from the Target service.

       Linkages are logged and emitted as INFO notifications.

       :params context: context object
       :params source_id: the Request ID of the source
       :params target_id: the Request ID of the target
       :params stage: optional event name extension to
                      indicate which part of the linkage
                      this is.
       :params target_name: human readable name of the
                            target system you are talking to.
       :params notifier: notifier object

       A typical use case is: System A asking System B
       to perform some action. The linkages might look
       like this:

       link_request_ids(sys_A.request_ID, stage="start")
       # send request to System B and get request ID
       link_request_ids(sys_A.request_ID, target_id=sys_B.request.ID)
       # optionally wait for System B to complete
       link_request_ids(sys_A.request_ID, target_id=sys_B.request.ID,
                        stage="end")

       But, it could be as simple as:
       link_request_ids(sys_A.request_ID, target_id=sys_B.request.ID)
       """

    event_name = "request.link"
    if stage:
        event_name += ".%s" % stage

    rtarget_id = ""
    if target_id:
        rtarget_id = _("TargetId=%(id)s ") % {'id': target_id}

    rtarget_name = ""
    if target_name:
        rtarget_name = _("Target='%(name)s' ") % {'name': target_name}

    arrow = ""
    if target_name or target_id:
        arrow = " -> "

    LOG.info(_("Request ID Link: %(event_name)s '%(source_id)s'%(arrow)s"
               "%(target_name)s%(target_id)s") % {"event_name": event_name,
                                                  "source_id": source_id,
                                                  "target_name": rtarget_name,
                                                  "arrow": arrow,
                                                  "target_id": rtarget_id})

    if notifier:
        payload = {"source_request_id": source_id,
                   "target_request_id": target_id,
                   "target_name": target_name,
                   "stage": stage}
        notifier.info(context, event_name, payload)



# Copyright (c) 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for the GlusterFS driver module."""

import contextlib
import errno
import mock
import os
import tempfile

import mox as mox_lib
from mox import IgnoreArg
from mox import IsA
from mox import stubout

from cinder import brick
from cinder import context
from cinder import db
from cinder import exception
from cinder.image import image_utils
from cinder.openstack.common import imageutils
from cinder.openstack.common import processutils as putils
from cinder import test
from cinder import units
from cinder import utils
from cinder.volume import configuration as conf
from cinder.volume import driver as base_driver
from cinder.volume.drivers import glusterfs


class DumbVolume(object):
    fields = {}

    def __setitem__(self, key, value):
        self.fields[key] = value

    def __getitem__(self, item):
        return self.fields[item]


class FakeDb(object):
    msg = "Tests are broken: mock this out."

    def volume_get(self, *a, **kw):
        raise Exception(self.msg)

    def snapshot_get_all_for_volume(self, *a, **kw):
        """Mock this if you want results from it."""
        return []


class GlusterFsDriverTestCase(test.TestCase):
    """Test case for GlusterFS driver."""

    TEST_EXPORT1 = 'glusterfs-host1:/export'
    TEST_EXPORT2 = 'glusterfs-host2:/export'
    TEST_EXPORT2_OPTIONS = '-o backupvolfile-server=glusterfs-backup1'
    TEST_SIZE_IN_GB = 1
    TEST_MNT_POINT = '/mnt/glusterfs'
    TEST_MNT_POINT_BASE = '/mnt/test'
    TEST_LOCAL_PATH = '/mnt/glusterfs/volume-123'
    TEST_FILE_NAME = 'test.txt'
    TEST_SHARES_CONFIG_FILE = '/etc/cinder/test-shares.conf'
    VOLUME_UUID = 'abcdefab-cdef-abcd-efab-cdefabcdefab'
    VOLUME_NAME = 'volume-%s' % VOLUME_UUID
    SNAP_UUID = 'bacadaca-baca-daca-baca-dacadacadaca'
    SNAP_UUID_2 = 'bebedede-bebe-dede-bebe-dedebebedede'

    def setUp(self):
        super(GlusterFsDriverTestCase, self).setUp()
        self._mox = mox_lib.Mox()
        self._configuration = mox_lib.MockObject(conf.Configuration)
        self._configuration.append_config_values(mox_lib.IgnoreArg())
        self._configuration.glusterfs_shares_config = \
            self.TEST_SHARES_CONFIG_FILE
        self._configuration.glusterfs_mount_point_base = \
            self.TEST_MNT_POINT_BASE
        self._configuration.glusterfs_disk_util = 'df'
        self._configuration.glusterfs_sparsed_volumes = True
        self._configuration.glusterfs_qcow2_volumes = False

        self.stubs = stubout.StubOutForTesting()
        self._driver =\
            glusterfs.GlusterfsDriver(configuration=self._configuration,
                                      db=FakeDb())
        self._driver.shares = {}

    def tearDown(self):
        self._mox.UnsetStubs()
        self.stubs.UnsetAll()
        super(GlusterFsDriverTestCase, self).tearDown()

    def stub_out_not_replaying(self, obj, attr_name):
        attr_to_replace = getattr(obj, attr_name)
        stub = mox_lib.MockObject(attr_to_replace)
        self.stubs.Set(obj, attr_name, stub)

    def test_set_execute(self):
        mox = self._mox
        drv = self._driver

        rfsclient = brick.remotefs.remotefs.RemoteFsClient

        mox.StubOutWithMock(rfsclient, 'set_execute')

        def my_execute(*a, **k):
            pass

        rfsclient.set_execute(my_execute)

        mox.ReplayAll()

        drv.set_execute(my_execute)

    def test_local_path(self):
        """local_path common use case."""
        glusterfs.CONF.glusterfs_mount_point_base = self.TEST_MNT_POINT_BASE
        drv = self._driver

        volume = DumbVolume()
        volume['provider_location'] = self.TEST_EXPORT1
        volume['name'] = 'volume-123'

        self.assertEqual(
            '/mnt/test/ab03ab34eaca46a5fb81878f7e9b91fc/volume-123',
            drv.local_path(volume))

    def test_mount_glusterfs_should_mount_correctly(self):
        """_mount_glusterfs common case usage."""
        mox = self._mox
        drv = self._driver

        mox.StubOutWithMock(drv, '_execute')
        drv._execute('mkdir', '-p', self.TEST_MNT_POINT)
        drv._execute('mount', '-t', 'glusterfs', self.TEST_EXPORT1,
                     self.TEST_MNT_POINT, run_as_root=True)

        mox.ReplayAll()

        drv._mount_glusterfs(self.TEST_EXPORT1, self.TEST_MNT_POINT)

        mox.VerifyAll()

    def test_mount_glusterfs_should_suppress_already_mounted_error(self):
        """_mount_glusterfs should suppress already mounted error if
           ensure=True
        """
        mox = self._mox
        drv = self._driver

        mox.StubOutWithMock(drv, '_execute')
        drv._execute('mkdir', '-p', self.TEST_MNT_POINT)
        drv._execute('mount', '-t', 'glusterfs', self.TEST_EXPORT1,
                     self.TEST_MNT_POINT, run_as_root=True).\
            AndRaise(putils.ProcessExecutionError(
                     stderr='is busy or already mounted'))

        mox.ReplayAll()

        drv._mount_glusterfs(self.TEST_EXPORT1, self.TEST_MNT_POINT,
                             ensure=True)

        mox.VerifyAll()

    def test_mount_glusterfs_should_reraise_already_mounted_error(self):
        """_mount_glusterfs should not suppress already mounted error
           if ensure=False
        """
        mox = self._mox
        drv = self._driver

        mox.StubOutWithMock(drv, '_execute')
        drv._execute('mkdir', '-p', self.TEST_MNT_POINT)
        drv._execute(
            'mount',
            '-t',
            'glusterfs',
            self.TEST_EXPORT1,
            self.TEST_MNT_POINT,
            run_as_root=True). \
            AndRaise(putils.ProcessExecutionError(stderr='is busy or '
                                                         'already mounted'))

        mox.ReplayAll()

        self.assertRaises(putils.ProcessExecutionError, drv._mount_glusterfs,
                          self.TEST_EXPORT1, self.TEST_MNT_POINT,
                          ensure=False)

        mox.VerifyAll()

    def test_mount_glusterfs_should_create_mountpoint_if_not_yet(self):
        """_mount_glusterfs should create mountpoint if it doesn't exist."""
        mox = self._mox
        drv = self._driver

        mox.StubOutWithMock(drv, '_execute')
        drv._execute('mkdir', '-p', self.TEST_MNT_POINT)
        drv._execute(*([IgnoreArg()] * 5), run_as_root=IgnoreArg())

        mox.ReplayAll()

        drv._mount_glusterfs(self.TEST_EXPORT1, self.TEST_MNT_POINT)

        mox.VerifyAll()

    def test_get_hash_str(self):
        """_get_hash_str should calculation correct value."""
        drv = self._driver

        self.assertEqual('ab03ab34eaca46a5fb81878f7e9b91fc',
                         drv._get_hash_str(self.TEST_EXPORT1))

    def test_get_mount_point_for_share(self):
        """_get_mount_point_for_share should call RemoteFsClient."""
        mox = self._mox
        drv = self._driver
        hashed_path = '/mnt/test/abcdefabcdef'

        mox.StubOutWithMock(brick.remotefs.remotefs.RemoteFsClient,
                            'get_mount_point')

        glusterfs.CONF.glusterfs_mount_point_base = self.TEST_MNT_POINT_BASE

        brick.remotefs.remotefs.RemoteFsClient.\
            get_mount_point(self.TEST_EXPORT1).AndReturn(hashed_path)

        mox.ReplayAll()

        drv._get_mount_point_for_share(self.TEST_EXPORT1)

    def test_get_available_capacity_with_df(self):
        """_get_available_capacity should calculate correct value."""
        mox = self._mox
        drv = self._driver

        df_total_size = 2620544
        df_avail = 1490560
        df_head = 'Filesystem 1K-blocks Used Available Use% Mounted on\n'
        df_data = 'glusterfs-host:/export %d 996864 %d 41%% /mnt' % \
                  (df_total_size, df_avail)
        df_output = df_head + df_data

        setattr(glusterfs.CONF, 'glusterfs_disk_util', 'df')

        mox.StubOutWithMock(drv, '_get_mount_point_for_share')
        drv._get_mount_point_for_share(self.TEST_EXPORT1).\
            AndReturn(self.TEST_MNT_POINT)

        mox.StubOutWithMock(drv, '_execute')
        drv._execute('df', '--portability', '--block-size', '1',
                     self.TEST_MNT_POINT,
                     run_as_root=True).AndReturn((df_output, None))

        mox.ReplayAll()

        self.assertEqual((df_avail, df_total_size),
                         drv._get_available_capacity(self.TEST_EXPORT1))

        mox.VerifyAll()

        delattr(glusterfs.CONF, 'glusterfs_disk_util')

    def test_load_shares_config(self):
        mox = self._mox
        drv = self._driver

        drv.configuration.glusterfs_shares_config = (
            self.TEST_SHARES_CONFIG_FILE)

        mox.StubOutWithMock(drv, '_read_config_file')
        config_data = []
        config_data.append(self.TEST_EXPORT1)
        config_data.append('#' + self.TEST_EXPORT2)
        config_data.append(self.TEST_EXPORT2 + ' ' + self.TEST_EXPORT2_OPTIONS)
        config_data.append('broken:share_format')
        config_data.append('')
        drv._read_config_file(self.TEST_SHARES_CONFIG_FILE).\
            AndReturn(config_data)
        mox.ReplayAll()

        drv._load_shares_config(drv.configuration.glusterfs_shares_config)

        self.assertIn(self.TEST_EXPORT1, drv.shares)
        self.assertIn(self.TEST_EXPORT2, drv.shares)
        self.assertEqual(len(drv.shares), 2)

        self.assertEqual(drv.shares[self.TEST_EXPORT2],
                         self.TEST_EXPORT2_OPTIONS)

        mox.VerifyAll()

    def test_ensure_share_mounted(self):
        """_ensure_share_mounted simple use case."""
        mox = self._mox
        drv = self._driver

        mox.StubOutWithMock(utils, 'get_file_mode')
        mox.StubOutWithMock(utils, 'get_file_gid')
        mox.StubOutWithMock(drv, '_execute')
        mox.StubOutWithMock(drv, '_ensure_share_writable')

        mox.StubOutWithMock(drv, '_get_mount_point_for_share')
        drv._get_mount_point_for_share(self.TEST_EXPORT1).\
            AndReturn(self.TEST_MNT_POINT)

        mox.StubOutWithMock(drv, '_mount_glusterfs')
        drv._mount_glusterfs(self.TEST_EXPORT1, self.TEST_MNT_POINT,
                             ensure=True)

        utils.get_file_gid(self.TEST_MNT_POINT).AndReturn(333333)

        utils.get_file_mode(self.TEST_MNT_POINT).AndReturn(0o777)

        drv._ensure_share_writable(self.TEST_MNT_POINT)

        drv._execute('chgrp', IgnoreArg(), self.TEST_MNT_POINT,
                     run_as_root=True)

        mox.ReplayAll()

        drv._ensure_share_mounted(self.TEST_EXPORT1)

        mox.VerifyAll()

    def test_ensure_shares_mounted_should_save_mounting_successfully(self):
        """_ensure_shares_mounted should save share if mounted with success."""
        mox = self._mox
        drv = self._driver

        mox.StubOutWithMock(drv, '_read_config_file')
        config_data = []
        config_data.append(self.TEST_EXPORT1)
        drv._read_config_file(self.TEST_SHARES_CONFIG_FILE).\
            AndReturn(config_data)

        mox.StubOutWithMock(drv, '_ensure_share_mounted')
        drv._ensure_share_mounted(self.TEST_EXPORT1)

        mox.ReplayAll()

        drv._ensure_shares_mounted()

        self.assertEqual(1, len(drv._mounted_shares))
        self.assertEqual(self.TEST_EXPORT1, drv._mounted_shares[0])

        mox.VerifyAll()

    def test_ensure_shares_mounted_should_not_save_mounting_with_error(self):
        """_ensure_shares_mounted should not save share if failed to mount."""
        mox = self._mox
        drv = self._driver

        mox.StubOutWithMock(drv, '_read_config_file')
        config_data = []
        config_data.append(self.TEST_EXPORT1)
        drv._read_config_file(self.TEST_SHARES_CONFIG_FILE).\
            AndReturn(config_data)

        mox.StubOutWithMock(drv, '_ensure_share_mounted')
        drv._ensure_share_mounted(self.TEST_EXPORT1).AndRaise(Exception())

        mox.ReplayAll()

        drv._ensure_shares_mounted()

        self.assertEqual(0, len(drv._mounted_shares))

        mox.VerifyAll()

    def test_setup_should_throw_error_if_shares_config_not_configured(self):
        """do_setup should throw error if shares config is not configured."""
        drv = self._driver

        glusterfs.CONF.glusterfs_shares_config = self.TEST_SHARES_CONFIG_FILE

        self.assertRaises(exception.GlusterfsException,
                          drv.do_setup, IsA(context.RequestContext))

    def test_setup_should_throw_exception_if_client_is_not_installed(self):
        """do_setup should throw exception if client is not installed."""
        mox = self._mox
        drv = self._driver

        glusterfs.CONF.glusterfs_shares_config = self.TEST_SHARES_CONFIG_FILE

        mox.StubOutWithMock(os.path, 'exists')
        os.path.exists(self.TEST_SHARES_CONFIG_FILE).AndReturn(True)
        mox.StubOutWithMock(drv, '_execute')
        drv._execute('mount.glusterfs', check_exit_code=False).\
            AndRaise(OSError(errno.ENOENT, 'No such file or directory'))

        mox.ReplayAll()

        self.assertRaises(exception.GlusterfsException,
                          drv.do_setup, IsA(context.RequestContext))

        mox.VerifyAll()

    def _fake_load_shares_config(self, conf):
        self._driver.shares = {'127.7.7.7:/gluster1': None}

    def _fake_NamedTemporaryFile(self, prefix=None, dir=None):
        raise OSError('Permission denied!')

    def test_setup_set_share_permissions(self):
        mox = self._mox
        drv = self._driver

        glusterfs.CONF.glusterfs_shares_config = self.TEST_SHARES_CONFIG_FILE

        self.stubs.Set(drv, '_load_shares_config',
                       self._fake_load_shares_config)
        self.stubs.Set(tempfile, 'NamedTemporaryFile',
                       self._fake_NamedTemporaryFile)
        mox.StubOutWithMock(os.path, 'exists')
        mox.StubOutWithMock(drv, '_execute')
        mox.StubOutWithMock(utils, 'get_file_gid')
        mox.StubOutWithMock(utils, 'get_file_mode')
        mox.StubOutWithMock(os, 'getegid')

        drv._execute('mount.glusterfs', check_exit_code=False)

        drv._execute('mkdir', '-p', mox_lib.IgnoreArg())

        os.path.exists(self.TEST_SHARES_CONFIG_FILE).AndReturn(True)

        drv._execute('mount', '-t', 'glusterfs', '127.7.7.7:/gluster1',
                     mox_lib.IgnoreArg(), run_as_root=True)

        utils.get_file_gid(mox_lib.IgnoreArg()).AndReturn(33333)
        # perms not writable
        utils.get_file_mode(mox_lib.IgnoreArg()).AndReturn(0o000)

        os.getegid().AndReturn(888)

        drv._execute('chgrp', 888, mox_lib.IgnoreArg(), run_as_root=True)
        drv._execute('chmod', 'g+w', mox_lib.IgnoreArg(), run_as_root=True)

        mox.ReplayAll()

        drv.do_setup(IsA(context.RequestContext))

        mox.VerifyAll()

    def test_find_share_should_throw_error_if_there_is_no_mounted_shares(self):
        """_find_share should throw error if there is no mounted shares."""
        drv = self._driver

        drv._mounted_shares = []

        self.assertRaises(exception.GlusterfsNoSharesMounted,
                          drv._find_share,
                          self.TEST_SIZE_IN_GB)

    def test_find_share(self):
        """_find_share simple use case."""
        mox = self._mox
        drv = self._driver

        drv._mounted_shares = [self.TEST_EXPORT1, self.TEST_EXPORT2]

        mox.StubOutWithMock(drv, '_get_available_capacity')
        drv._get_available_capacity(self.TEST_EXPORT1).\
            AndReturn((2 * units.GiB, 5 * units.GiB))
        drv._get_available_capacity(self.TEST_EXPORT2).\
            AndReturn((3 * units.GiB, 10 * units.GiB))

        mox.ReplayAll()

        self.assertEqual(self.TEST_EXPORT2,
                         drv._find_share(self.TEST_SIZE_IN_GB))

        mox.VerifyAll()

    def test_find_share_should_throw_error_if_there_is_no_enough_place(self):
        """_find_share should throw error if there is no share to host vol."""
        mox = self._mox
        drv = self._driver

        drv._mounted_shares = [self.TEST_EXPORT1,
                               self.TEST_EXPORT2]

        mox.StubOutWithMock(drv, '_get_available_capacity')
        drv._get_available_capacity(self.TEST_EXPORT1).\
            AndReturn((0, 5 * units.GiB))
        drv._get_available_capacity(self.TEST_EXPORT2).\
            AndReturn((0, 10 * units.GiB))

        mox.ReplayAll()

        self.assertRaises(exception.GlusterfsNoSuitableShareFound,
                          drv._find_share,
                          self.TEST_SIZE_IN_GB)

        mox.VerifyAll()

    def _simple_volume(self, id=None):
        volume = DumbVolume()
        volume['provider_location'] = self.TEST_EXPORT1
        if id is None:
            volume['id'] = self.VOLUME_UUID
        else:
            volume['id'] = id
        # volume['name'] mirrors format from db/sqlalchemy/models.py
        volume['name'] = 'volume-%s' % volume['id']
        volume['size'] = 10
        volume['status'] = 'available'

        return volume

    def test_create_sparsed_volume(self):
        mox = self._mox
        drv = self._driver
        volume = self._simple_volume()

        setattr(glusterfs.CONF, 'glusterfs_sparsed_volumes', True)

        mox.StubOutWithMock(drv, '_create_sparsed_file')
        mox.StubOutWithMock(drv, '_set_rw_permissions_for_all')

        drv._create_sparsed_file(IgnoreArg(), IgnoreArg())
        drv._set_rw_permissions_for_all(IgnoreArg())

        mox.ReplayAll()

        drv._do_create_volume(volume)

        mox.VerifyAll()

        delattr(glusterfs.CONF, 'glusterfs_sparsed_volumes')

    def test_create_nonsparsed_volume(self):
        mox = self._mox
        drv = self._driver
        volume = self._simple_volume()

        old_value = self._configuration.glusterfs_sparsed_volumes
        self._configuration.glusterfs_sparsed_volumes = False

        mox.StubOutWithMock(drv, '_create_regular_file')
        mox.StubOutWithMock(drv, '_set_rw_permissions_for_all')

        drv._create_regular_file(IgnoreArg(), IgnoreArg())
        drv._set_rw_permissions_for_all(IgnoreArg())

        mox.ReplayAll()

        drv._do_create_volume(volume)

        mox.VerifyAll()

        self._configuration.glusterfs_sparsed_volumes = old_value

    def test_create_qcow2_volume(self):
        (mox, drv) = self._mox, self._driver

        volume = self._simple_volume()
        old_value = self._configuration.glusterfs_qcow2_volumes
        self._configuration.glusterfs_qcow2_volumes = True

        mox.StubOutWithMock(drv, '_execute')

        hashed = drv._get_hash_str(volume['provider_location'])
        path = '%s/%s/volume-%s' % (self.TEST_MNT_POINT_BASE,
                                    hashed,
                                    self.VOLUME_UUID)

        drv._execute('qemu-img', 'create', '-f', 'qcow2',
                     '-o', 'preallocation=metadata', path,
                     str(volume['size'] * units.GiB),
                     run_as_root=True)

        drv._execute('chmod', 'ugo+rw', path, run_as_root=True)

        mox.ReplayAll()

        drv._do_create_volume(volume)

        mox.VerifyAll()

        self._configuration.glusterfs_qcow2_volumes = old_value

    def test_create_volume_should_ensure_glusterfs_mounted(self):
        """create_volume ensures shares provided in config are mounted."""
        mox = self._mox
        drv = self._driver

        self.stub_out_not_replaying(glusterfs, 'LOG')
        self.stub_out_not_replaying(drv, '_find_share')
        self.stub_out_not_replaying(drv, '_do_create_volume')

        mox.StubOutWithMock(drv, '_ensure_shares_mounted')
        drv._ensure_shares_mounted()

        mox.ReplayAll()

        volume = DumbVolume()
        volume['size'] = self.TEST_SIZE_IN_GB
        drv.create_volume(volume)

        mox.VerifyAll()

    def test_create_volume_should_return_provider_location(self):
        """create_volume should return provider_location with found share."""
        mox = self._mox
        drv = self._driver

        self.stub_out_not_replaying(glusterfs, 'LOG')
        self.stub_out_not_replaying(drv, '_ensure_shares_mounted')
        self.stub_out_not_replaying(drv, '_do_create_volume')

        mox.StubOutWithMock(drv, '_find_share')
        drv._find_share(self.TEST_SIZE_IN_GB).AndReturn(self.TEST_EXPORT1)

        mox.ReplayAll()

        volume = DumbVolume()
        volume['size'] = self.TEST_SIZE_IN_GB
        result = drv.create_volume(volume)
        self.assertEqual(self.TEST_EXPORT1, result['provider_location'])

        mox.VerifyAll()

    def test_create_cloned_volume(self):
        (mox, drv) = self._mox, self._driver

        mox.StubOutWithMock(drv, '_create_snapshot')
        mox.StubOutWithMock(drv, '_delete_snapshot')
        mox.StubOutWithMock(drv, '_read_info_file')
        mox.StubOutWithMock(image_utils, 'convert_image')
        mox.StubOutWithMock(drv, '_copy_volume_from_snapshot')

        volume_file = 'volume-%s' % self.VOLUME_UUID
        volume_path = '%s/%s/%s' % (self.TEST_MNT_POINT_BASE,
                                    drv._get_hash_str(self.TEST_EXPORT1),
                                    volume_file)

        volume = self._simple_volume()
        src_vref = self._simple_volume()
        src_vref['id'] = '375e32b2-804a-49f2-b282-85d1d5a5b9e1'
        src_vref['name'] = 'volume-%s' % src_vref['id']
        volume_file = 'volume-%s' % src_vref['id']
        volume_path = '%s/%s/%s' % (self.TEST_MNT_POINT_BASE,
                                    drv._get_hash_str(self.TEST_EXPORT1),
                                    volume_file)
        src_info_path = '%s.info' % volume_path
        volume_ref = {'id': volume['id'],
                      'name': volume['name'],
                      'status': volume['status'],
                      'provider_location': volume['provider_location'],
                      'size': volume['size']}

        snap_ref = {'volume_name': src_vref['name'],
                    'name': 'clone-snap-%s' % src_vref['id'],
                    'size': src_vref['size'],
                    'volume_size': src_vref['size'],
                    'volume_id': src_vref['id'],
                    'id': 'tmp-snap-%s' % src_vref['id'],
                    'volume': src_vref}

        drv._create_snapshot(snap_ref)

        snap_info = {'active': volume_file,
                     snap_ref['id']: volume_path + '-clone'}

        drv._read_info_file(src_info_path).AndReturn(snap_info)

        drv._copy_volume_from_snapshot(snap_ref, volume_ref, volume['size'])

        drv._delete_snapshot(mox_lib.IgnoreArg())

        mox.ReplayAll()

        drv.create_cloned_volume(volume, src_vref)

    @mock.patch('cinder.openstack.common.fileutils.delete_if_exists')
    def test_delete_volume(self, mock_delete_if_exists):
        volume = self._simple_volume()
        volume_filename = 'volume-%s' % self.VOLUME_UUID
        volume_path = '%s/%s' % (self.TEST_MNT_POINT, volume_filename)
        info_file = volume_path + '.info'

        with contextlib.nested(
                mock.patch.object(self._driver, '_ensure_share_mounted'),
                mock.patch.object(self._driver, '_local_volume_dir'),
                mock.patch.object(self._driver, 'get_active_image_from_info'),
                mock.patch.object(self._driver, '_execute'),
                mock.patch.object(self._driver, '_local_path_volume_info')
        ) as (mock_ensure_share_mounted, mock_local_volume_dir,
              mock_active_image_from_info, mock_execute,
              mock_local_path_volume_info):
            mock_local_volume_dir.return_value = self.TEST_MNT_POINT
            mock_active_image_from_info.return_value = volume_filename
            mock_local_path_volume_info.return_value = info_file

            self._driver.delete_volume(volume)

            mock_ensure_share_mounted.assert_called_once_with(
                volume['provider_location'])
            mock_local_volume_dir.assert_called_once_with(volume)
            mock_active_image_from_info.assert_called_once_with(volume)
            mock_execute.assert_called_once_with('rm', '-f', volume_path,
                                                 run_as_root=True)
            mock_local_path_volume_info.assert_called_once_with(volume)
            mock_delete_if_exists.assert_called_once_with(info_file)

    def test_delete_should_ensure_share_mounted(self):
        """delete_volume should ensure that corresponding share is mounted."""
        mox = self._mox
        drv = self._driver

        self.stub_out_not_replaying(drv, '_execute')

        volume = DumbVolume()
        volume['name'] = 'volume-123'
        volume['provider_location'] = self.TEST_EXPORT1

        mox.StubOutWithMock(drv, '_ensure_share_mounted')
        drv._ensure_share_mounted(self.TEST_EXPORT1)

        mox.ReplayAll()

        drv.delete_volume(volume)

        mox.VerifyAll()

    def test_delete_should_not_delete_if_provider_location_not_provided(self):
        """delete_volume shouldn't delete if provider_location missed."""
        mox = self._mox
        drv = self._driver

        self.stub_out_not_replaying(drv, '_ensure_share_mounted')

        volume = DumbVolume()
        volume['name'] = 'volume-123'
        volume['provider_location'] = None

        mox.StubOutWithMock(drv, '_execute')

        mox.ReplayAll()

        drv.delete_volume(volume)

        mox.VerifyAll()

    def test_create_snapshot(self):
        (mox, drv) = self._mox, self._driver

        self.stub_out_not_replaying(drv, '_ensure_share_mounted')
        mox.StubOutWithMock(drv, '_create_qcow2_snap_file')
        mox.StubOutWithMock(drv, '_read_info_file')
        mox.StubOutWithMock(drv, '_write_info_file')

        volume = self._simple_volume()
        snap_ref = {'name': 'test snap',
                    'volume_id': self.VOLUME_UUID,
                    'volume': volume,
                    'id': self.SNAP_UUID}

        mox.StubOutWithMock(drv, '_execute')

        vol_filename = 'volume-%s' % self.VOLUME_UUID
        snap_filename = '%s.%s' % (vol_filename, self.SNAP_UUID)

        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        vol_path = '%s/%s/%s' % (self.TEST_MNT_POINT_BASE,
                                 hashed,
                                 vol_filename)
        snap_path = '%s.%s' % (vol_path, self.SNAP_UUID)
        info_path = '%s%s' % (vol_path, '.info')

        info_dict = {'active': vol_filename}
        drv._read_info_file(info_path, empty_if_missing=True).\
            AndReturn(info_dict)

        drv._create_qcow2_snap_file(snap_ref, vol_filename, snap_path)

        qemu_img_info_output = ("""image: volume-%s
        file format: raw
        virtual size: 1.0G (1073741824 bytes)
        disk size: 152K
        """ % self.VOLUME_UUID, '')

        drv._read_info_file(info_path, empty_if_missing=True).\
            AndReturn(info_dict)

        # SNAP_UUID_2 has been removed from dict.
        info_file_dict = {'active': 'volume-%s.%s' %
                          (self.VOLUME_UUID, self.SNAP_UUID),
                          self.SNAP_UUID: 'volume-%s.%s' %
                          (self.VOLUME_UUID, self.SNAP_UUID)}

        drv._write_info_file(info_path, info_file_dict)

        mox.ReplayAll()

        drv.create_snapshot(snap_ref)

        mox.VerifyAll()

    def test_delete_snapshot_bottom(self):
        """Multiple snapshots exist.

           In this test, path (volume-<uuid>) is backed by
            snap_path (volume-<uuid>.<snap_uuid>) which is backed by
            snap_path_2 (volume-<uuid>.<snap_uuid_2>).

           Delete the snapshot identified by SNAP_UUID_2.

           Chain goes from
                               (SNAP_UUID)      (SNAP_UUID_2)
             volume-abc -> volume-abc.baca -> volume-abc.bebe
           to
                               (SNAP_UUID)
             volume-abc -> volume-abc.baca
        """
        (mox, drv) = self._mox, self._driver

        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        volume_dir = os.path.join(self.TEST_MNT_POINT_BASE, hashed)
        volume_path = '%s/%s/volume-%s' % (self.TEST_MNT_POINT_BASE,
                                           hashed,
                                           self.VOLUME_UUID)
        volume_filename = 'volume-%s' % self.VOLUME_UUID

        snap_path = '%s.%s' % (volume_path, self.SNAP_UUID)
        snap_path_2 = '%s.%s' % (volume_path, self.SNAP_UUID_2)
        snap_file = '%s.%s' % (volume_filename, self.SNAP_UUID)
        snap_file_2 = '%s.%s' % (volume_filename, self.SNAP_UUID_2)
        info_path = '%s%s' % (volume_path, '.info')

        qemu_img_info_output = """image: volume-%s.%s
        file format: qcow2
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        backing file: %s
        """ % (self.VOLUME_UUID, self.SNAP_UUID, volume_filename)

        mox.StubOutWithMock(drv, '_execute')
        mox.StubOutWithMock(drv, '_read_file')
        mox.StubOutWithMock(drv, '_read_info_file')
        mox.StubOutWithMock(drv, '_get_backing_chain_for_path')
        mox.StubOutWithMock(drv, '_get_matching_backing_file')
        mox.StubOutWithMock(drv, '_write_info_file')
        mox.StubOutWithMock(drv, '_ensure_share_writable')
        mox.StubOutWithMock(image_utils, 'qemu_img_info')

        drv._ensure_share_writable(volume_dir)

        img_info = imageutils.QemuImgInfo(qemu_img_info_output)
        image_utils.qemu_img_info(snap_path_2).AndReturn(img_info)

        info_file_dict = {'active': snap_file_2,
                          self.SNAP_UUID_2: snap_file_2,
                          self.SNAP_UUID: snap_file}

        snap_ref = {'name': 'test snap',
                    'volume_id': self.VOLUME_UUID,
                    'volume': self._simple_volume(),
                    'id': self.SNAP_UUID_2}

        snap_path_2_chain = [{self.SNAP_UUID_2: snap_file_2},
                             {self.SNAP_UUID: snap_file},
                             {'active': snap_file_2}]

        snap_path_chain = [{self.SNAP_UUID: snap_file},
                           {'active': snap_file}]

        drv._read_info_file(info_path, empty_if_missing=True).\
            AndReturn(info_file_dict)

        drv._execute('qemu-img', 'commit', snap_path_2, run_as_root=True)

        drv._execute('rm', '-f', snap_path_2, run_as_root=True)

        drv._read_info_file(info_path, empty_if_missing=True).\
            AndReturn(info_file_dict)

        drv._read_info_file(info_path).AndReturn(info_file_dict)

        drv._write_info_file(info_path, info_file_dict)

        mox.ReplayAll()

        drv.delete_snapshot(snap_ref)

        mox.VerifyAll()

    def test_delete_snapshot_middle(self):
        """Multiple snapshots exist.

           In this test, path (volume-<uuid>) is backed by
            snap_path (volume-<uuid>.<snap_uuid>) which is backed by
            snap_path_2 (volume-<uuid>.<snap_uuid_2>).

           Delete the snapshot identified with SNAP_UUID.

           Chain goes from
                               (SNAP_UUID)      (SNAP_UUID_2)
             volume-abc -> volume-abc.baca -> volume-abc.bebe
           to                (SNAP_UUID_2)
             volume-abc -> volume-abc.bebe
        """
        (mox, drv) = self._mox, self._driver

        volume = self._simple_volume()

        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        volume_file = 'volume-%s' % self.VOLUME_UUID
        volume_dir = os.path.join(self.TEST_MNT_POINT_BASE, hashed)
        volume_path = '%s/%s/%s' % (self.TEST_MNT_POINT_BASE,
                                    hashed,
                                    volume_file)

        info_path = '%s%s' % (volume_path, '.info')
        snap_path = '%s.%s' % (volume_path, self.SNAP_UUID)
        snap_file = 'volume-%s.%s' % (self.VOLUME_UUID, self.SNAP_UUID)
        snap_path_2 = '%s.%s' % (volume_path, self.SNAP_UUID_2)
        snap_file_2 = 'volume-%s.%s' % (self.VOLUME_UUID, self.SNAP_UUID_2)

        qemu_img_info_output_snap_2 = """image: volume-%s.%s
        file format: qcow2
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        backing file: %s
        """ % (self.VOLUME_UUID, self.SNAP_UUID_2,
               'volume-%s.%s' % (self.VOLUME_UUID, self.SNAP_UUID_2))

        qemu_img_info_output_snap_1 = """image: volume-%s.%s
        file format: qcow2
        virtual size: 1.0G (1073741824 bytes)
        disk size: 122K
        backing file: %s
        """ % (self.VOLUME_UUID, self.SNAP_UUID,
               'volume-%s.%s' % (self.VOLUME_UUID, self.SNAP_UUID))

        qemu_img_info_output = """image: volume-%s
        file format: qcow2
        virtual size: 1.0G (1073741824 bytes)
        disk size: 175K
        """ % self.VOLUME_UUID

        mox.StubOutWithMock(drv, '_execute')
        mox.StubOutWithMock(drv, '_read_info_file')
        mox.StubOutWithMock(drv, '_write_info_file')
        mox.StubOutWithMock(drv, '_get_backing_chain_for_path')
        mox.StubOutWithMock(drv, 'get_active_image_from_info')
        mox.StubOutWithMock(drv, '_ensure_share_writable')
        mox.StubOutWithMock(image_utils, 'qemu_img_info')

        info_file_dict = {self.SNAP_UUID_2: 'volume-%s.%s' %
                          (self.VOLUME_UUID, self.SNAP_UUID_2),
                          self.SNAP_UUID: 'volume-%s.%s' %
                          (self.VOLUME_UUID, self.SNAP_UUID)}

        drv._ensure_share_writable(volume_dir)

        info_path = drv._local_path_volume(volume) + '.info'
        drv._read_info_file(info_path, empty_if_missing=True).\
            AndReturn(info_file_dict)

        img_info = imageutils.QemuImgInfo(qemu_img_info_output_snap_1)
        image_utils.qemu_img_info(snap_path).AndReturn(img_info)

        snap_ref = {'name': 'test snap',
                    'volume_id': self.VOLUME_UUID,
                    'volume': volume,
                    'id': self.SNAP_UUID}

        snap_path_chain = [{'filename': snap_file_2,
                            'backing-filename': snap_file},
                           {'filename': snap_file,
                            'backing-filename': volume_file}]

        drv.get_active_image_from_info(volume).AndReturn(snap_file_2)
        drv._get_backing_chain_for_path(volume, snap_path_2).\
            AndReturn(snap_path_chain)

        drv._read_info_file(info_path).AndReturn(info_file_dict)

        drv._execute('qemu-img', 'commit', snap_path_2, run_as_root=True)

        drv._execute('rm', '-f', snap_path_2, run_as_root=True)

        drv._read_info_file(info_path).AndReturn(info_file_dict)

        drv._write_info_file(info_path, info_file_dict)

        mox.ReplayAll()

        drv.delete_snapshot(snap_ref)

        mox.VerifyAll()

    def test_delete_snapshot_not_in_info(self):
        """Snapshot not in info file / info file doesn't exist.

        Snapshot creation failed so nothing is on-disk.  Driver
        should allow operation to succeed so the manager can
        remove the snapshot record.

        (Scenario: Snapshot object created in Cinder db but not
         on backing storage.)

        """
        (mox, drv) = self._mox, self._driver

        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        volume_dir = os.path.join(self.TEST_MNT_POINT_BASE, hashed)
        volume_filename = 'volume-%s' % self.VOLUME_UUID
        volume_path = os.path.join(volume_dir, volume_filename)
        info_path = '%s%s' % (volume_path, '.info')

        mox.StubOutWithMock(drv, '_read_file')
        mox.StubOutWithMock(drv, '_read_info_file')
        mox.StubOutWithMock(drv, '_ensure_share_writable')

        snap_ref = {'name': 'test snap',
                    'volume_id': self.VOLUME_UUID,
                    'volume': self._simple_volume(),
                    'id': self.SNAP_UUID_2}

        drv._ensure_share_writable(volume_dir)

        drv._read_info_file(info_path, empty_if_missing=True).AndReturn({})

        mox.ReplayAll()

        drv.delete_snapshot(snap_ref)

        mox.VerifyAll()

    def test_read_info_file(self):
        (mox, drv) = self._mox, self._driver

        mox.StubOutWithMock(drv, '_read_file')
        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        volume_path = '%s/%s/volume-%s' % (self.TEST_MNT_POINT_BASE,
                                           hashed,
                                           self.VOLUME_UUID)
        info_path = '%s%s' % (volume_path, '.info')

        drv._read_file(info_path).AndReturn('{"%(id)s": "volume-%(id)s"}' %
                                            {'id': self.VOLUME_UUID})

        mox.ReplayAll()

        volume = DumbVolume()
        volume['id'] = self.VOLUME_UUID
        volume['name'] = 'volume-%s' % self.VOLUME_UUID

        info = drv._read_info_file(info_path)

        self.assertEqual(info[self.VOLUME_UUID],
                         'volume-%s' % self.VOLUME_UUID)

        mox.VerifyAll()

    def test_extend_volume(self):
        (mox, drv) = self._mox, self._driver

        volume = self._simple_volume()

        volume_path = '%s/%s/volume-%s' % (self.TEST_MNT_POINT_BASE,
                                           drv._get_hash_str(
                                               self.TEST_EXPORT1),
                                           self.VOLUME_UUID)

        qemu_img_info_output = """image: volume-%s
        file format: qcow2
        virtual size: 1.0G (1073741824 bytes)
        disk size: 473K
        """ % self.VOLUME_UUID

        img_info = imageutils.QemuImgInfo(qemu_img_info_output)

        mox.StubOutWithMock(drv, '_execute')
        mox.StubOutWithMock(drv, 'get_active_image_from_info')
        mox.StubOutWithMock(image_utils, 'qemu_img_info')
        mox.StubOutWithMock(image_utils, 'resize_image')

        drv.get_active_image_from_info(volume).AndReturn(volume['name'])

        image_utils.qemu_img_info(volume_path).AndReturn(img_info)

        image_utils.resize_image(volume_path, 3)

        mox.ReplayAll()

        drv.extend_volume(volume, 3)

        mox.VerifyAll()

    def test_create_snapshot_online(self):
        (mox, drv) = self._mox, self._driver

        volume = self._simple_volume()
        volume['status'] = 'in-use'

        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        volume_file = 'volume-%s' % self.VOLUME_UUID
        volume_path = '%s/%s/%s' % (self.TEST_MNT_POINT_BASE,
                                    hashed,
                                    volume_file)
        info_path = '%s.info' % volume_path

        ctxt = context.RequestContext('fake_user', 'fake_project')

        snap_ref = {'name': 'test snap (online)',
                    'volume_id': self.VOLUME_UUID,
                    'volume': volume,
                    'id': self.SNAP_UUID,
                    'context': ctxt,
                    'status': 'asdf',
                    'progress': 'asdf'}

        snap_path = '%s.%s' % (volume_path, self.SNAP_UUID)
        snap_file = '%s.%s' % (volume_file, self.SNAP_UUID)

        mox.StubOutWithMock(drv, '_execute')
        mox.StubOutWithMock(drv, '_create_qcow2_snap_file')
        mox.StubOutWithMock(db, 'snapshot_get')
        mox.StubOutWithMock(drv, '_write_info_file')
        mox.StubOutWithMock(drv, '_nova')

        drv._create_qcow2_snap_file(snap_ref, volume_file, snap_path)

        create_info = {'snapshot_id': snap_ref['id'],
                       'type': 'qcow2',
                       'new_file': snap_file}

        drv._nova.create_volume_snapshot(ctxt, self.VOLUME_UUID, create_info)

        snap_ref['status'] = 'creating'
        snap_ref['progress'] = '0%'
        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        snap_ref['progress'] = '50%'
        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        snap_ref['progress'] = '90%'
        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        snap_info = {'active': snap_file,
                     self.SNAP_UUID: snap_file}

        drv._write_info_file(info_path, snap_info)

        mox.ReplayAll()

        drv.create_snapshot(snap_ref)

    def test_create_snapshot_online_novafailure(self):
        (mox, drv) = self._mox, self._driver

        volume = self._simple_volume()
        volume['status'] = 'in-use'

        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        volume_file = 'volume-%s' % self.VOLUME_UUID
        volume_path = '%s/%s/%s' % (self.TEST_MNT_POINT_BASE,
                                    hashed,
                                    volume_file)
        info_path = '%s.info' % volume_path

        ctxt = context.RequestContext('fake_user', 'fake_project')

        snap_ref = {'name': 'test snap (online)',
                    'volume_id': self.VOLUME_UUID,
                    'volume': volume,
                    'id': self.SNAP_UUID,
                    'context': ctxt}

        snap_path = '%s.%s' % (volume_path, self.SNAP_UUID)
        snap_file = '%s.%s' % (volume_file, self.SNAP_UUID)

        mox.StubOutWithMock(drv, '_execute')
        mox.StubOutWithMock(drv, '_create_qcow2_snap_file')
        mox.StubOutWithMock(drv, '_nova')
        mox.StubOutWithMock(db, 'snapshot_get')
        mox.StubOutWithMock(drv, '_write_info_file')

        drv._create_qcow2_snap_file(snap_ref, volume_file, snap_path)

        create_info = {'snapshot_id': snap_ref['id'],
                       'type': 'qcow2',
                       'new_file': snap_file}

        drv._nova.create_volume_snapshot(ctxt, self.VOLUME_UUID, create_info)

        snap_ref['status'] = 'creating'
        snap_ref['progress'] = '0%'

        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        snap_ref['progress'] = '50%'
        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        snap_ref['progress'] = '99%'
        snap_ref['status'] = 'error'
        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        snap_info = {'active': snap_file,
                     self.SNAP_UUID: snap_file}

        drv._write_info_file(info_path, snap_info)

        mox.ReplayAll()

        self.assertRaises(exception.GlusterfsException,
                          drv.create_snapshot,
                          snap_ref)

    def test_delete_snapshot_online_1(self):
        """Delete the newest snapshot, with only one snap present."""
        (mox, drv) = self._mox, self._driver

        volume = self._simple_volume()
        volume['status'] = 'in-use'

        ctxt = context.RequestContext('fake_user', 'fake_project')

        snap_ref = {'name': 'test snap to delete (online)',
                    'volume_id': self.VOLUME_UUID,
                    'volume': volume,
                    'id': self.SNAP_UUID,
                    'context': ctxt}

        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        volume_file = 'volume-%s' % self.VOLUME_UUID
        volume_dir = os.path.join(self.TEST_MNT_POINT_BASE, hashed)
        volume_path = '%s/%s/%s' % (self.TEST_MNT_POINT_BASE,
                                    hashed,
                                    volume_file)
        info_path = '%s.info' % volume_path

        snap_path = '%s.%s' % (volume_path, self.SNAP_UUID)
        snap_file = '%s.%s' % (volume_file, self.SNAP_UUID)

        mox.StubOutWithMock(drv, '_execute')
        mox.StubOutWithMock(drv, '_nova')
        mox.StubOutWithMock(drv, '_read_info_file')
        mox.StubOutWithMock(drv, '_write_info_file')
        mox.StubOutWithMock(os.path, 'exists')
        mox.StubOutWithMock(db, 'snapshot_get')
        mox.StubOutWithMock(image_utils, 'qemu_img_info')
        mox.StubOutWithMock(drv, '_ensure_share_writable')

        snap_info = {'active': snap_file,
                     self.SNAP_UUID: snap_file}

        drv._ensure_share_writable(volume_dir)

        drv._read_info_file(info_path, empty_if_missing=True).\
            AndReturn(snap_info)

        os.path.exists(snap_path).AndReturn(True)

        qemu_img_info_output = """image: %s
        file format: qcow2
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        backing file: %s
        """ % (snap_file, volume_file)
        img_info = imageutils.QemuImgInfo(qemu_img_info_output)

        vol_qemu_img_info_output = """image: %s
        file format: raw
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        """ % volume_file
        volume_img_info = imageutils.QemuImgInfo(vol_qemu_img_info_output)

        image_utils.qemu_img_info(snap_path).AndReturn(img_info)

        image_utils.qemu_img_info(volume_path).AndReturn(volume_img_info)

        drv._read_info_file(info_path, empty_if_missing=True).\
            AndReturn(snap_info)

        delete_info = {
            'type': 'qcow2',
            'merge_target_file': None,
            'file_to_merge': None,
            'volume_id': self.VOLUME_UUID
        }

        drv._nova.delete_volume_snapshot(ctxt, self.SNAP_UUID, delete_info)

        drv._read_info_file(info_path).AndReturn(snap_info)

        drv._read_info_file(info_path).AndReturn(snap_info)

        snap_ref['status'] = 'deleting'
        snap_ref['progress'] = '0%'
        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        snap_ref['progress'] = '50%'
        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        snap_ref['progress'] = '90%'
        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        drv._write_info_file(info_path, snap_info)

        drv._execute('rm', '-f', volume_path, run_as_root=True)

        mox.ReplayAll()

        drv.delete_snapshot(snap_ref)

    def test_delete_snapshot_online_2(self):
        """Delete the middle of 3 snapshots."""
        (mox, drv) = self._mox, self._driver

        volume = self._simple_volume()
        volume['status'] = 'in-use'

        ctxt = context.RequestContext('fake_user', 'fake_project')

        snap_ref = {'name': 'test snap to delete (online)',
                    'volume_id': self.VOLUME_UUID,
                    'volume': volume,
                    'id': self.SNAP_UUID,
                    'context': ctxt}

        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        volume_file = 'volume-%s' % self.VOLUME_UUID
        volume_dir = os.path.join(self.TEST_MNT_POINT_BASE, hashed)
        volume_path = '%s/%s/%s' % (self.TEST_MNT_POINT_BASE,
                                    hashed,
                                    volume_file)
        info_path = '%s.info' % volume_path

        snap_path = '%s.%s' % (volume_path, self.SNAP_UUID)
        snap_path_2 = '%s.%s' % (volume_path, self.SNAP_UUID_2)
        snap_file = '%s.%s' % (volume_file, self.SNAP_UUID)
        snap_file_2 = '%s.%s' % (volume_file, self.SNAP_UUID_2)

        mox.StubOutWithMock(drv, '_execute')
        mox.StubOutWithMock(drv, '_nova')
        mox.StubOutWithMock(drv, '_read_info_file')
        mox.StubOutWithMock(drv, '_write_info_file')
        mox.StubOutWithMock(os.path, 'exists')
        mox.StubOutWithMock(db, 'snapshot_get')
        mox.StubOutWithMock(image_utils, 'qemu_img_info')
        mox.StubOutWithMock(drv, '_ensure_share_writable')

        snap_info = {'active': snap_file_2,
                     self.SNAP_UUID: snap_file,
                     self.SNAP_UUID_2: snap_file_2}

        drv._ensure_share_writable(volume_dir)

        drv._read_info_file(info_path, empty_if_missing=True).\
            AndReturn(snap_info)

        os.path.exists(snap_path).AndReturn(True)

        qemu_img_info_output = """image: %s
        file format: qcow2
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        backing file: %s
        """ % (snap_file, volume_file)
        img_info = imageutils.QemuImgInfo(qemu_img_info_output)

        vol_qemu_img_info_output = """image: %s
        file format: raw
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        """ % volume_file
        volume_img_info = imageutils.QemuImgInfo(vol_qemu_img_info_output)

        image_utils.qemu_img_info(snap_path).AndReturn(img_info)

        image_utils.qemu_img_info(volume_path).AndReturn(volume_img_info)

        drv._read_info_file(info_path, empty_if_missing=True).\
            AndReturn(snap_info)

        delete_info = {'type': 'qcow2',
                       'merge_target_file': volume_file,
                       'file_to_merge': snap_file,
                       'volume_id': self.VOLUME_UUID}
        drv._nova.delete_volume_snapshot(ctxt, self.SNAP_UUID, delete_info)

        drv._read_info_file(info_path).AndReturn(snap_info)

        drv._read_info_file(info_path).AndReturn(snap_info)

        snap_ref['status'] = 'deleting'
        snap_ref['progress'] = '0%'
        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        snap_ref['progress'] = '50%'
        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        snap_ref['progress'] = '90%'
        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        drv._write_info_file(info_path, snap_info)

        drv._execute('rm', '-f', snap_path, run_as_root=True)

        mox.ReplayAll()

        drv.delete_snapshot(snap_ref)

    def test_delete_snapshot_online_novafailure(self):
        """Delete the newest snapshot."""
        (mox, drv) = self._mox, self._driver

        volume = self._simple_volume()
        volume['status'] = 'in-use'

        ctxt = context.RequestContext('fake_user', 'fake_project')

        snap_ref = {'name': 'test snap to delete (online)',
                    'volume_id': self.VOLUME_UUID,
                    'volume': volume,
                    'id': self.SNAP_UUID,
                    'context': ctxt}

        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        volume_file = 'volume-%s' % self.VOLUME_UUID
        volume_path = '%s/%s/%s' % (self.TEST_MNT_POINT_BASE,
                                    hashed,
                                    volume_file)
        info_path = '%s.info' % volume_path

        snap_path = '%s.%s' % (volume_path, self.SNAP_UUID)
        snap_file = '%s.%s' % (volume_file, self.SNAP_UUID)

        mox.StubOutWithMock(drv, '_execute')
        mox.StubOutWithMock(drv, '_nova')
        mox.StubOutWithMock(drv, '_read_info_file')
        mox.StubOutWithMock(drv, '_write_info_file')
        mox.StubOutWithMock(os.path, 'exists')
        mox.StubOutWithMock(db, 'snapshot_get')
        mox.StubOutWithMock(image_utils, 'qemu_img_info')

        snap_info = {'active': snap_file,
                     self.SNAP_UUID: snap_file}

        drv._read_info_file(info_path, empty_if_missing=True).\
            AndReturn(snap_info)

        os.path.exists(snap_path).AndReturn(True)

        qemu_img_info_output = """image: %s
        file format: qcow2
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        backing file: %s
        """ % (snap_file, volume_file)
        img_info = imageutils.QemuImgInfo(qemu_img_info_output)

        image_utils.qemu_img_info(snap_path).AndReturn(img_info)

        drv._read_info_file(info_path, empty_if_missing=True).\
            AndReturn(snap_info)

        delete_info = {
            'type': 'qcow2',
            'merge_target_file': None,
            'file_to_merge': volume_file,
            'volume_id': self.VOLUME_UUID
        }

        drv._nova.delete_volume_snapshot(ctxt, self.SNAP_UUID, delete_info)

        drv._read_info_file(info_path).AndReturn(snap_info)

        drv._read_info_file(info_path).AndReturn(snap_info)

        snap_ref['status'] = 'deleting'
        snap_ref['progress'] = '0%'
        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        snap_ref['progress'] = '50%'
        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        snap_ref['status'] = 'error_deleting'
        snap_ref['progress'] = '90%'
        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        db.snapshot_get(ctxt, self.SNAP_UUID).AndReturn(snap_ref)

        drv._write_info_file(info_path, snap_info)

        drv._execute('rm', '-f', volume_path, run_as_root=True)

        mox.ReplayAll()

        self.assertRaises(exception.GlusterfsException,
                          drv.delete_snapshot,
                          snap_ref)

    def test_get_backing_chain_for_path(self):
        (mox, drv) = self._mox, self._driver

        glusterfs.CONF.glusterfs_mount_point_base = self.TEST_MNT_POINT_BASE

        volume = self._simple_volume()
        vol_filename = volume['name']
        vol_filename_2 = volume['name'] + '.abcd'
        vol_filename_3 = volume['name'] + '.efef'
        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        vol_dir = '%s/%s' % (self.TEST_MNT_POINT_BASE, hashed)
        vol_path = '%s/%s' % (vol_dir, vol_filename)
        vol_path_2 = '%s/%s' % (vol_dir, vol_filename_2)
        vol_path_3 = '%s/%s' % (vol_dir, vol_filename_3)

        mox.StubOutWithMock(drv, '_execute')
        mox.StubOutWithMock(drv, '_local_volume_dir')
        mox.StubOutWithMock(image_utils, 'qemu_img_info')

        qemu_img_output_base = """image: %(image_name)s
        file format: qcow2
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        """
        qemu_img_output = """image: %(image_name)s
        file format: qcow2
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        backing file: %(backing_file)s
        """

        qemu_img_output_1 = qemu_img_output_base % {'image_name': vol_filename}
        qemu_img_output_2 = qemu_img_output % {'image_name': vol_filename_2,
                                               'backing_file': vol_filename}
        qemu_img_output_3 = qemu_img_output % {'image_name': vol_filename_3,
                                               'backing_file': vol_filename_2}

        info_1 = imageutils.QemuImgInfo(qemu_img_output_1)
        info_2 = imageutils.QemuImgInfo(qemu_img_output_2)
        info_3 = imageutils.QemuImgInfo(qemu_img_output_3)

        drv._local_volume_dir(volume).AndReturn(vol_dir)
        image_utils.qemu_img_info(vol_path_3).\
            AndReturn(info_3)
        drv._local_volume_dir(volume).AndReturn(vol_dir)
        image_utils.qemu_img_info(vol_path_2).\
            AndReturn(info_2)
        drv._local_volume_dir(volume).AndReturn(vol_dir)
        image_utils.qemu_img_info(vol_path).\
            AndReturn(info_1)

        mox.ReplayAll()

        chain = drv._get_backing_chain_for_path(volume, vol_path_3)

        # Verify chain contains all expected data
        item_1 = drv._get_matching_backing_file(chain, vol_filename)
        self.assertEqual(item_1['filename'], vol_filename_2)
        chain.remove(item_1)
        item_2 = drv._get_matching_backing_file(chain, vol_filename_2)
        self.assertEqual(item_2['filename'], vol_filename_3)
        chain.remove(item_2)
        self.assertEqual(len(chain), 1)
        self.assertEqual(chain[0]['filename'], vol_filename)

    def test_copy_volume_from_snapshot(self):
        (mox, drv) = self._mox, self._driver

        mox.StubOutWithMock(image_utils, 'convert_image')
        mox.StubOutWithMock(drv, '_read_info_file')
        mox.StubOutWithMock(image_utils, 'qemu_img_info')
        mox.StubOutWithMock(drv, '_set_rw_permissions_for_all')

        dest_volume = self._simple_volume(
            'c1073000-0000-0000-0000-0000000c1073')
        src_volume = self._simple_volume()

        vol_dir = os.path.join(self.TEST_MNT_POINT_BASE,
                               drv._get_hash_str(self.TEST_EXPORT1))
        src_vol_path = os.path.join(vol_dir, src_volume['name'])
        dest_vol_path = os.path.join(vol_dir, dest_volume['name'])
        info_path = os.path.join(vol_dir, src_volume['name']) + '.info'

        snapshot = {'volume_name': src_volume['name'],
                    'name': 'clone-snap-%s' % src_volume['id'],
                    'size': src_volume['size'],
                    'volume_size': src_volume['size'],
                    'volume_id': src_volume['id'],
                    'id': 'tmp-snap-%s' % src_volume['id'],
                    'volume': src_volume}

        snap_file = dest_volume['name'] + '.' + snapshot['id']
        snap_path = os.path.join(vol_dir, snap_file)

        size = dest_volume['size']

        drv._read_info_file(info_path).AndReturn(
            {'active': snap_file,
             snapshot['id']: snap_file}
        )

        qemu_img_output = """image: %s
        file format: raw
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        backing file: %s
        """ % (snap_file, src_volume['name'])
        img_info = imageutils.QemuImgInfo(qemu_img_output)

        image_utils.qemu_img_info(snap_path).AndReturn(img_info)

        image_utils.convert_image(src_vol_path, dest_vol_path, 'raw')

        drv._set_rw_permissions_for_all(dest_vol_path)

        mox.ReplayAll()

        drv._copy_volume_from_snapshot(snapshot, dest_volume, size)

    def test_create_volume_from_snapshot(self):
        (mox, drv) = self._mox, self._driver

        volume = self._simple_volume('c1073000-0000-0000-0000-0000000c1073')
        src_volume = self._simple_volume()

        mox.StubOutWithMock(drv, '_create_snapshot')
        mox.StubOutWithMock(drv, '_copy_volume_from_snapshot')
        mox.StubOutWithMock(drv, '_delete_snapshot')

        snap_ref = {'volume_name': src_volume['name'],
                    'name': 'clone-snap-%s' % src_volume['id'],
                    'size': src_volume['size'],
                    'volume_size': src_volume['size'],
                    'volume_id': src_volume['id'],
                    'id': 'tmp-snap-%s' % src_volume['id'],
                    'volume': src_volume}

        volume_ref = {'id': volume['id'],
                      'size': volume['size'],
                      'status': volume['status'],
                      'provider_location': volume['provider_location'],
                      'name': 'volume-' + volume['id']}

        drv._create_snapshot(snap_ref)
        drv._copy_volume_from_snapshot(snap_ref,
                                       volume_ref,
                                       src_volume['size'])
        drv._delete_snapshot(snap_ref)

        mox.ReplayAll()

        drv.create_cloned_volume(volume, src_volume)

    def test_initialize_connection(self):
        (mox, drv) = self._mox, self._driver

        volume = self._simple_volume()
        vol_dir = os.path.join(self.TEST_MNT_POINT_BASE,
                               drv._get_hash_str(self.TEST_EXPORT1))
        vol_path = os.path.join(vol_dir, volume['name'])

        qemu_img_output = """image: %s
        file format: raw
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        """ % volume['name']
        img_info = imageutils.QemuImgInfo(qemu_img_output)

        mox.StubOutWithMock(drv, 'get_active_image_from_info')
        mox.StubOutWithMock(image_utils, 'qemu_img_info')

        drv.get_active_image_from_info(volume).AndReturn(volume['name'])
        image_utils.qemu_img_info(vol_path).AndReturn(img_info)

        mox.ReplayAll()

        conn_info = drv.initialize_connection(volume, None)

        self.assertEqual(conn_info['data']['format'], 'raw')
        self.assertEqual(conn_info['driver_volume_type'], 'glusterfs')
        self.assertEqual(conn_info['data']['name'], volume['name'])
        self.assertEqual(conn_info['mount_point_base'],
                         self.TEST_MNT_POINT_BASE)

    def test_get_mount_point_base(self):
        (mox, drv) = self._mox, self._driver

        self.assertEqual(drv._get_mount_point_base(),
                         self.TEST_MNT_POINT_BASE)

    def test_backup_volume(self):
        """Backup a volume with no snapshots."""

        (mox, drv) = self._mox, self._driver

        mox.StubOutWithMock(drv, '_qemu_img_info')
        mox.StubOutWithMock(drv.db, 'volume_get')
        mox.StubOutWithMock(base_driver.VolumeDriver, 'backup_volume')
        mox.StubOutWithMock(drv, '_read_info_file')
        mox.StubOutWithMock(drv, 'get_active_image_from_info')

        ctxt = context.RequestContext('fake_user', 'fake_project')
        volume = self._simple_volume()
        backup = {'volume_id': volume['id']}

        drv._read_info_file(IgnoreArg(), empty_if_missing=True).AndReturn({})
        drv.get_active_image_from_info(IgnoreArg()).AndReturn('/some/path')

        info = imageutils.QemuImgInfo()
        info.file_format = 'raw'

        drv.db.volume_get(ctxt, volume['id']).AndReturn(volume)
        drv._qemu_img_info(IgnoreArg(), IgnoreArg()).AndReturn(info)

        base_driver.VolumeDriver.backup_volume(IgnoreArg(),
                                               IgnoreArg(),
                                               IgnoreArg())

        mox.ReplayAll()

        drv.backup_volume(ctxt, backup, IgnoreArg())

    def test_backup_volume_previous_snap(self):
        """Backup a volume that previously had a snapshot.

           Snapshot was deleted, snap_info is different from above.
        """

        (mox, drv) = self._mox, self._driver

        mox.StubOutWithMock(drv, '_qemu_img_info')
        mox.StubOutWithMock(drv.db, 'volume_get')
        mox.StubOutWithMock(drv, '_read_info_file')
        mox.StubOutWithMock(drv, 'get_active_image_from_info')
        mox.StubOutWithMock(base_driver.VolumeDriver, 'backup_volume')

        ctxt = context.RequestContext('fake_user', 'fake_project')
        volume = self._simple_volume()
        backup = {'volume_id': volume['id']}

        drv._read_info_file(IgnoreArg(), empty_if_missing=True).AndReturn(
            {'active': 'file2'})
        drv.get_active_image_from_info(IgnoreArg()).AndReturn('/some/file2')

        info = imageutils.QemuImgInfo()
        info.file_format = 'raw'

        drv.db.volume_get(ctxt, volume['id']).AndReturn(volume)
        drv._qemu_img_info(IgnoreArg(), IgnoreArg()).AndReturn(info)

        base_driver.VolumeDriver.backup_volume(IgnoreArg(),
                                               IgnoreArg(),
                                               IgnoreArg())

        mox.ReplayAll()

        drv.backup_volume(ctxt, backup, IgnoreArg())

    def test_backup_snap_failure_1(self):
        """Backup fails if snapshot exists (database)."""

        (mox, drv) = self._mox, self._driver
        mox.StubOutWithMock(drv.db, 'snapshot_get_all_for_volume')
        mox.StubOutWithMock(base_driver.VolumeDriver, 'backup_volume')

        ctxt = context.RequestContext('fake_user', 'fake_project')
        volume = self._simple_volume()
        backup = {'volume_id': volume['id']}

        drv.db.snapshot_get_all_for_volume(ctxt, volume['id']).AndReturn(
            [{'snap1': 'a'}, {'snap2': 'b'}])

        base_driver.VolumeDriver.backup_volume(IgnoreArg(),
                                               IgnoreArg(),
                                               IgnoreArg())

        mox.ReplayAll()

        self.assertRaises(exception.InvalidVolume,
                          drv.backup_volume,
                          ctxt, backup, IgnoreArg())

    def test_backup_snap_failure_2(self):
        """Backup fails if snapshot exists (on-disk)."""

        (mox, drv) = self._mox, self._driver
        mox.StubOutWithMock(drv, '_read_info_file')
        mox.StubOutWithMock(drv.db, 'volume_get')
        mox.StubOutWithMock(drv, 'get_active_image_from_info')
        mox.StubOutWithMock(drv, '_qemu_img_info')
        mox.StubOutWithMock(base_driver.VolumeDriver, 'backup_volume')

        ctxt = context.RequestContext('fake_user', 'fake_project')
        volume = self._simple_volume()
        backup = {'volume_id': volume['id']}

        drv.db.volume_get(ctxt, volume['id']).AndReturn(volume)

        drv._read_info_file(IgnoreArg(), empty_if_missing=True).AndReturn(
            {'id1': 'file1',
             'id2': 'file2',
             'active': 'file2'})

        drv.get_active_image_from_info(IgnoreArg()).\
            AndReturn('/some/path/file2')

        info = imageutils.QemuImgInfo()
        info.file_format = 'raw'
        info.backing_file = 'file1'

        drv._qemu_img_info(IgnoreArg(), IgnoreArg()).AndReturn(info)

        base_driver.VolumeDriver.backup_volume(IgnoreArg(),
                                               IgnoreArg(),
                                               IgnoreArg())

        mox.ReplayAll()

        self.assertRaises(exception.InvalidVolume,
                          drv.backup_volume,
                          ctxt, backup, IgnoreArg())

    def test_backup_failure_unsupported_format(self):
        """Attempt to backup a volume with a qcow2 base."""

        (mox, drv) = self._mox, self._driver

        mox.StubOutWithMock(drv, '_qemu_img_info')
        mox.StubOutWithMock(drv.db, 'volume_get')
        mox.StubOutWithMock(base_driver.VolumeDriver, 'backup_volume')
        mox.StubOutWithMock(drv, '_read_info_file')
        mox.StubOutWithMock(drv, 'get_active_image_from_info')

        ctxt = context.RequestContext('fake_user', 'fake_project')
        volume = self._simple_volume()
        backup = {'volume_id': volume['id']}

        drv._read_info_file(IgnoreArg(), empty_if_missing=True).AndReturn({})
        drv.get_active_image_from_info(IgnoreArg()).AndReturn('/some/path')

        info = imageutils.QemuImgInfo()
        info.file_format = 'qcow2'

        drv.db.volume_get(ctxt, volume['id']).AndReturn(volume)
        drv._qemu_img_info(IgnoreArg(), IgnoreArg()).AndReturn(info)

        base_driver.VolumeDriver.backup_volume(IgnoreArg(),
                                               IgnoreArg(),
                                               IgnoreArg())

        mox.ReplayAll()

        self.assertRaises(exception.InvalidVolume,
                          drv.backup_volume,
                          ctxt, backup, IgnoreArg())



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Driver for Linux servers running LVM.

"""

import math
import os
import socket

from oslo.config import cfg

from cinder.brick import exception as brick_exception
from cinder.brick.local_dev import lvm as lvm
from cinder import exception
from cinder.image import image_utils
from cinder.openstack.common import fileutils
from cinder.openstack.common import log as logging
from cinder.openstack.common import processutils
from cinder import units
from cinder import utils
from cinder.volume import driver
from cinder.volume import utils as volutils

LOG = logging.getLogger(__name__)

volume_opts = [
    cfg.StrOpt('volume_group',
               default='cinder-volumes',
               help='Name for the VG that will contain exported volumes'),
    cfg.IntOpt('lvm_mirrors',
               default=0,
               help='If set, create lvms with multiple mirrors. Note that '
                    'this requires lvm_mirrors + 2 pvs with available space'),
    cfg.StrOpt('lvm_type',
               default='default',
               help='Type of LVM volumes to deploy; (default or thin)'),
]

CONF = cfg.CONF
CONF.register_opts(volume_opts)


class LVMVolumeDriver(driver.VolumeDriver):
    """Executes commands relating to Volumes."""

    VERSION = '2.0.0'

    def __init__(self, vg_obj=None, *args, **kwargs):
        super(LVMVolumeDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(volume_opts)
        self.hostname = socket.gethostname()
        self.vg = vg_obj
        self.backend_name =\
            self.configuration.safe_get('volume_backend_name') or 'LVM'
        self.protocol = 'local'

    def set_execute(self, execute):
        self._execute = execute

    def check_for_setup_error(self):
        """Verify that requirements are in place to use LVM driver."""
        if self.vg is None:
            root_helper = utils.get_root_helper()
            try:
                self.vg = lvm.LVM(self.configuration.volume_group,
                                  root_helper,
                                  lvm_type=self.configuration.lvm_type,
                                  executor=self._execute)
            except brick_exception.VolumeGroupNotFound:
                message = ("Volume Group %s does not exist" %
                           self.configuration.volume_group)
                raise exception.VolumeBackendAPIException(data=message)

        vg_list = volutils.get_all_volume_groups(
            self.configuration.volume_group)
        vg_dict = \
            (vg for vg in vg_list if vg['name'] == self.vg.vg_name).next()
        if vg_dict is None:
            message = ("Volume Group %s does not exist" %
                       self.configuration.volume_group)
            raise exception.VolumeBackendAPIException(data=message)

        if self.configuration.lvm_type == 'thin':
            # Specific checks for using Thin provisioned LV's
            if not volutils.supports_thin_provisioning():
                message = ("Thin provisioning not supported "
                           "on this version of LVM.")
                raise exception.VolumeBackendAPIException(data=message)

            pool_name = "%s-pool" % self.configuration.volume_group
            if self.vg.get_volume(pool_name) is None:
                try:
                    self.vg.create_thin_pool(pool_name)
                except processutils.ProcessExecutionError as exc:
                    exception_message = ("Failed to create thin pool, "
                                         "error message was: %s"
                                         % exc.stderr)
                    raise exception.VolumeBackendAPIException(
                        data=exception_message)

    def _sizestr(self, size_in_g):
        if int(size_in_g) == 0:
            return '100m'
        return '%sg' % size_in_g

    def _volume_not_present(self, volume_name):
        return self.vg.get_volume(volume_name) is None

    def _delete_volume(self, volume, is_snapshot=False):
        """Deletes a logical volume."""
        if self.configuration.volume_clear != 'none' and \
                self.configuration.lvm_type != 'thin':
            self._clear_volume(volume, is_snapshot)

        name = volume['name']
        if is_snapshot:
            name = self._escape_snapshot(volume['name'])
        self.vg.delete(name)

    def _clear_volume(self, volume, is_snapshot=False):
        # zero out old volumes to prevent data leaking between users
        # TODO(ja): reclaiming space should be done lazy and low priority
        if is_snapshot:
            # if the volume to be cleared is a snapshot of another volume
            # we need to clear out the volume using the -cow instead of the
            # directly volume path.  We need to skip this if we are using
            # thin provisioned LVs.
            # bug# lp1191812
            dev_path = self.local_path(volume) + "-cow"
        else:
            dev_path = self.local_path(volume)

        # TODO(jdg): Maybe we could optimize this for snaps by looking at
        # the cow table and only overwriting what's necessary?
        # for now we're still skipping on snaps due to hang issue
        if not os.path.exists(dev_path):
            msg = (_('Volume device file path %s does not exist.')
                   % dev_path)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        size_in_g = volume.get('size', volume.get('volume_size', None))
        if size_in_g is None:
            msg = (_("Size for volume: %s not found, "
                     "cannot secure delete.") % volume['id'])
            LOG.error(msg)
            raise exception.InvalidParameterValue(msg)

        # clear_volume expects sizes in MiB, we store integer GiB
        # be sure to convert before passing in
        vol_sz_in_meg = size_in_g * units.KiB

        volutils.clear_volume(
            vol_sz_in_meg, dev_path,
            volume_clear=self.configuration.volume_clear,
            volume_clear_size=self.configuration.volume_clear_size)

    def _escape_snapshot(self, snapshot_name):
        # Linux LVM reserves name that starts with snapshot, so that
        # such volume name can't be created. Mangle it.
        if not snapshot_name.startswith('snapshot'):
            return snapshot_name
        return '_' + snapshot_name

    def _create_volume(self, name, size, lvm_type, mirror_count, vg=None):
        vg_ref = self.vg
        if vg is not None:
            vg_ref = vg

        vg_ref.create_volume(name, size, lvm_type, mirror_count)

    def create_volume(self, volume):
        """Creates a logical volume."""
        mirror_count = 0
        if self.configuration.lvm_mirrors:
            mirror_count = self.configuration.lvm_mirrors

        self._create_volume(volume['name'],
                            self._sizestr(volume['size']),
                            self.configuration.lvm_type,
                            mirror_count)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from a snapshot."""
        self._create_volume(volume['name'],
                            self._sizestr(volume['size']),
                            self.configuration.lvm_type,
                            self.configuration.lvm_mirrors)

        # Some configurations of LVM do not automatically activate
        # ThinLVM snapshot LVs.
        self.vg.activate_lv(snapshot['name'], is_snapshot=True)

       # copy_volume expects sizes in MiB, we store integer GiB
       # be sure to convert before passing in
        volutils.copy_volume(self.local_path(snapshot),
                             self.local_path(volume),
                             snapshot['volume_size'] * units.KiB,
                             self.configuration.volume_dd_blocksize,
                             execute=self._execute)

    def delete_volume(self, volume):
        """Deletes a logical volume."""

        # NOTE(jdg):  We don't need to explicitly call
        # remove export here because we already did it
        # in the manager before we got here.

        if self._volume_not_present(volume['name']):
            # If the volume isn't present, then don't attempt to delete
            return True

        if self.vg.lv_has_snapshot(volume['name']):
            LOG.error(_('Unabled to delete due to existing snapshot '
                        'for volume: %s') % volume['name'])
            raise exception.VolumeIsBusy(volume_name=volume['name'])

        self._delete_volume(volume)

    def create_snapshot(self, snapshot):
        """Creates a snapshot."""

        self.vg.create_lv_snapshot(self._escape_snapshot(snapshot['name']),
                                   snapshot['volume_name'],
                                   self.configuration.lvm_type)

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot."""
        if self._volume_not_present(self._escape_snapshot(snapshot['name'])):
            # If the snapshot isn't present, then don't attempt to delete
            LOG.warning(_("snapshot: %s not found, "
                          "skipping delete operations") % snapshot['name'])
            return True

        # TODO(yamahata): zeroing out the whole snapshot triggers COW.
        # it's quite slow.
        self._delete_volume(snapshot, is_snapshot=True)

    def local_path(self, volume, vg=None):
        if vg is None:
            vg = self.configuration.volume_group
        # NOTE(vish): stops deprecation warning
        escaped_group = vg.replace('-', '--')
        escaped_name = self._escape_snapshot(volume['name']).replace('-', '--')
        return "/dev/mapper/%s-%s" % (escaped_group, escaped_name)

    def copy_image_to_volume(self, context, volume, image_service, image_id):
        """Fetch the image from image_service and write it to the volume."""
        image_utils.fetch_to_raw(context,
                                 image_service,
                                 image_id,
                                 self.local_path(volume),
                                 self.configuration.volume_dd_blocksize,
                                 size=volume['size'])

    def copy_volume_to_image(self, context, volume, image_service, image_meta):
        """Copy the volume to the specified image."""
        image_utils.upload_volume(context,
                                  image_service,
                                  image_meta,
                                  self.local_path(volume))

    def create_cloned_volume(self, volume, src_vref):
        """Creates a clone of the specified volume."""

        mirror_count = 0
        if self.configuration.lvm_mirrors:
            mirror_count = self.configuration.lvm_mirrors
        LOG.info(_('Creating clone of volume: %s') % src_vref['id'])
        volume_name = src_vref['name']
        temp_id = 'tmp-snap-%s' % volume['id']
        temp_snapshot = {'volume_name': volume_name,
                         'size': src_vref['size'],
                         'volume_size': src_vref['size'],
                         'name': 'clone-snap-%s' % volume['id'],
                         'id': temp_id}

        self.create_snapshot(temp_snapshot)

       # copy_volume expects sizes in MiB, we store integer GiB
       # be sure to convert before passing in
        try:
            self._create_volume(volume['name'],
                                self._sizestr(volume['size']),
                                self.configuration.lvm_type,
                                mirror_count)

            self.vg.activate_lv(temp_snapshot['name'], is_snapshot=True)

            volutils.copy_volume(
                self.local_path(temp_snapshot),
                self.local_path(volume),
                src_vref['size'] * units.KiB,
                self.configuration.volume_dd_blocksize,
                execute=self._execute)
        finally:
            self.delete_snapshot(temp_snapshot)

    def clone_image(self, volume, image_location, image_id, image_meta):
        return None, False

    def backup_volume(self, context, backup, backup_service):
        """Create a new backup from an existing volume."""
        volume = self.db.volume_get(context, backup['volume_id'])
        volume_path = self.local_path(volume)
        with utils.temporary_chown(volume_path):
            with fileutils.file_open(volume_path) as volume_file:
                backup_service.backup(backup, volume_file)

    def restore_backup(self, context, backup, volume, backup_service):
        """Restore an existing backup to a new or existing volume."""
        volume_path = self.local_path(volume)
        with utils.temporary_chown(volume_path):
            with fileutils.file_open(volume_path, 'wb') as volume_file:
                backup_service.restore(backup, volume['id'], volume_file)

    def get_volume_stats(self, refresh=False):
        """Get volume status.

        If 'refresh' is True, run update the stats first.
        """

        if refresh:
            self._update_volume_stats()

        return self._stats

    def _update_volume_stats(self):
        """Retrieve stats info from volume group."""

        LOG.debug(_("Updating volume stats"))
        if self.vg is None:
            LOG.warning(_('Unable to update stats on non-initialized '
                          'Volume Group: %s'), self.configuration.volume_group)
            return

        self.vg.update_volume_group_info()
        data = {}

        # Note(zhiteng): These information are driver/backend specific,
        # each driver may define these values in its own config options
        # or fetch from driver specific configuration file.
        data["volume_backend_name"] = self.backend_name
        data["vendor_name"] = 'Open Source'
        data["driver_version"] = self.VERSION
        data["storage_protocol"] = self.protocol

        if self.configuration.lvm_mirrors > 0:
            data['total_capacity_gb'] =\
                self.vg.vg_mirror_size(self.configuration.lvm_mirrors)
            data['free_capacity_gb'] =\
                self.vg.vg_mirror_free_space(self.configuration.lvm_mirrors)
        elif self.configuration.lvm_type == 'thin':
            data['total_capacity_gb'] = self.vg.vg_thin_pool_size
            data['free_capacity_gb'] = self.vg.vg_thin_pool_free_space
        else:
            data['total_capacity_gb'] = self.vg.vg_size
            data['free_capacity_gb'] = self.vg.vg_free_space
        data['reserved_percentage'] = self.configuration.reserved_percentage
        data['QoS_support'] = False
        data['location_info'] =\
            ('LVMVolumeDriver:%(hostname)s:%(vg)s'
             ':%(lvm_type)s:%(lvm_mirrors)s' %
             {'hostname': self.hostname,
              'vg': self.configuration.volume_group,
              'lvm_type': self.configuration.lvm_type,
              'lvm_mirrors': self.configuration.lvm_mirrors})

        self._stats = data

    def extend_volume(self, volume, new_size):
        """Extend an existing volume's size."""
        self.vg.extend_volume(volume['name'],
                              self._sizestr(new_size))

    def manage_existing(self, volume, existing_ref):
        """Manages an existing LV.

        Renames the LV to match the expected name for the volume.
        Error checking done by manage_existing_get_size is not repeated.
        """
        lv_name = existing_ref['lv_name']
        lv = self.vg.get_volume(lv_name)

        # Attempt to rename the LV to match the OpenStack internal name.
        try:
            self.vg.rename_volume(lv_name, volume['name'])
        except processutils.ProcessExecutionError as exc:
            exception_message = (_("Failed to rename logical volume %(name)s, "
                                   "error message was: %(err_msg)s")
                                 % {'name': lv_name,
                                    'err_msg': exc.stderr})
            raise exception.VolumeBackendAPIException(
                data=exception_message)

    def manage_existing_get_size(self, volume, existing_ref):
        """Return size of an existing LV for manage_existing.

        existing_ref is a dictionary of the form:
        {'lv_name': <name of LV>}
        """

        # Check that the reference is valid
        if 'lv_name' not in existing_ref:
            reason = _('Reference must contain lv_name element.')
            raise exception.ManageExistingInvalidReference(
                existing_ref=existing_ref, reason=reason)
        lv_name = existing_ref['lv_name']
        lv = self.vg.get_volume(lv_name)

        # Raise an exception if we didn't find a suitable LV.
        if not lv:
            kwargs = {'existing_ref': lv_name,
                      'reason': 'Specified logical volume does not exist.'}
            raise exception.ManageExistingInvalidReference(**kwargs)

        # LV size is returned in gigabytes.  Attempt to parse size as a float
        # and round up to the next integer.
        try:
            lv_size = int(math.ceil(float(lv['size'])))
        except ValueError:
            exception_message = (_("Failed to manage existing volume "
                                   "%(name)s, because reported size %(size)s "
                                   "was not a floating-point number.")
                                 % {'name': lv_name,
                                    'size': lv['size']})
            raise exception.VolumeBackendAPIException(
                data=exception_message)
        return lv_size


class LVMISCSIDriver(LVMVolumeDriver, driver.ISCSIDriver):
    """Executes commands relating to ISCSI volumes.

    We make use of model provider properties as follows:

    ``provider_location``
      if present, contains the iSCSI target information in the same
      format as an ietadm discovery
      i.e. '<ip>:<port>,<portal> <target IQN>'

    ``provider_auth``
      if present, contains a space-separated triple:
      '<auth method> <auth username> <auth password>'.
      `CHAP` is the only auth_method in use at the moment.
    """

    def __init__(self, *args, **kwargs):
        self.db = kwargs.get('db')
        self.target_helper = self.get_target_helper(self.db)
        super(LVMISCSIDriver, self).__init__(*args, **kwargs)
        self.backend_name =\
            self.configuration.safe_get('volume_backend_name') or 'LVM_iSCSI'
        self.protocol = 'iSCSI'

    def set_execute(self, execute):
        super(LVMISCSIDriver, self).set_execute(execute)
        if self.target_helper is not None:
            self.target_helper.set_execute(execute)

    def _create_target(self, iscsi_name, iscsi_target,
                       volume_path, chap_auth, lun=0,
                       check_exit_code=False, old_name=None):
        # NOTE(jdg): tgt driver has an issue where with a lot of activity
        # (or sometimes just randomly) it will get *confused* and attempt
        # to reuse a target ID, resulting in a target already exists error
        # Typically a simple retry will address this

        # For now we have this while loop, might be useful in the
        # future to throw a retry decorator in common or utils
        attempts = 2
        while attempts > 0:
            attempts -= 1
            try:
                # NOTE(jdg): For TgtAdm case iscsi_name is all we need
                # should clean this all up at some point in the future
                tid = self.target_helper.create_iscsi_target(
                    iscsi_name,
                    iscsi_target,
                    0,
                    volume_path,
                    chap_auth,
                    check_exit_code=check_exit_code,
                    old_name=old_name)
                break

            except brick_exception.ISCSITargetCreateFailed:
                if attempts == 0:
                    raise
                else:
                    LOG.warning(_('Error creating iSCSI target, retrying '
                                  'creation for target: %s') % iscsi_name)
        return tid

    def ensure_export(self, context, volume):
        volume_name = volume['name']
        iscsi_name = "%s%s" % (self.configuration.iscsi_target_prefix,
                               volume_name)
        volume_path = "/dev/%s/%s" % (self.configuration.volume_group,
                                      volume_name)
        # NOTE(jdg): For TgtAdm case iscsi_name is the ONLY param we need
        # should clean this all up at some point in the future
        model_update = self.target_helper.ensure_export(context, volume,
                                                        iscsi_name,
                                                        volume_path)
        if model_update:
            self.db.volume_update(context, volume['id'], model_update)

    def create_export(self, context, volume):
        return self._create_export(context, volume)

    def _create_export(self, context, volume, vg=None):
        """Creates an export for a logical volume."""
        if vg is None:
            vg = self.configuration.volume_group

        volume_path = "/dev/%s/%s" % (vg, volume['name'])

        data = self.target_helper.create_export(context, volume, volume_path)
        return {
            'provider_location': data['location'],
            'provider_auth': data['auth'],
        }

    def remove_export(self, context, volume):
        self.target_helper.remove_export(context, volume)

    def migrate_volume(self, ctxt, volume, host, thin=False, mirror_count=0):
        """Optimize the migration if the destination is on the same server.

        If the specified host is another back-end on the same server, and
        the volume is not attached, we can do the migration locally without
        going through iSCSI.
        """

        false_ret = (False, None)
        if volume['status'] != 'available':
            return false_ret
        if 'location_info' not in host['capabilities']:
            return false_ret
        info = host['capabilities']['location_info']
        try:
            (dest_type, dest_hostname, dest_vg, lvm_type, lvm_mirrors) =\
                info.split(':')
            lvm_mirrors = int(lvm_mirrors)
        except ValueError:
            return false_ret
        if (dest_type != 'LVMVolumeDriver' or dest_hostname != self.hostname):
            return false_ret

        if dest_vg != self.vg.vg_name:
            vg_list = volutils.get_all_volume_groups()
            try:
                (vg for vg in vg_list if vg['name'] == dest_vg).next()
            except StopIteration:
                message = ("Destination Volume Group %s does not exist" %
                           dest_vg)
                LOG.error(_('%s'), message)
                return false_ret

            helper = utils.get_root_helper()
            dest_vg_ref = lvm.LVM(dest_vg, helper,
                                  lvm_type=lvm_type,
                                  executor=self._execute)
            self.remove_export(ctxt, volume)
            self._create_volume(volume['name'],
                                self._sizestr(volume['size']),
                                lvm_type,
                                lvm_mirrors,
                                dest_vg_ref)

        volutils.copy_volume(self.local_path(volume),
                             self.local_path(volume, vg=dest_vg),
                             volume['size'],
                             self.configuration.volume_dd_blocksize,
                             execute=self._execute)
        self._delete_volume(volume)
        model_update = self._create_export(ctxt, volume, vg=dest_vg)

        return (True, model_update)

    def _iscsi_location(self, ip, target, iqn, lun=None):
        return "%s:%s,%s %s %s" % (ip, self.configuration.iscsi_port,
                                   target, iqn, lun)

    def _iscsi_authentication(self, chap, name, password):
        return "%s %s %s" % (chap, name, password)


class LVMISERDriver(LVMISCSIDriver, driver.ISERDriver):
    """Executes commands relating to ISER volumes.

    We make use of model provider properties as follows:

    ``provider_location``
      if present, contains the iSER target information in the same
      format as an ietadm discovery
      i.e. '<ip>:<port>,<portal> <target IQN>'

    ``provider_auth``
      if present, contains a space-separated triple:
      '<auth method> <auth username> <auth password>'.
      `CHAP` is the only auth_method in use at the moment.
    """

    def __init__(self, *args, **kwargs):
        self.target_helper = self.get_target_helper(kwargs.get('db'))
        LVMVolumeDriver.__init__(self, *args, **kwargs)
        self.backend_name =\
            self.configuration.safe_get('volume_backend_name') or 'LVM_iSER'
        self.protocol = 'iSER'



# (c) Copyright 2013 OpenStack Foundation
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import mox

from cinder.brick import exception
from cinder.brick.remotefs import remotefs
from cinder.openstack.common import log as logging
from cinder import test

LOG = logging.getLogger(__name__)


class BrickRemoteFsTestCase(test.TestCase):
    TEST_EXPORT = '1.2.3.4/export1'
    TEST_MNT_BASE = '/mnt/test'
    TEST_HASH = '4d664fd43b6ff86d80a4ea969c07b3b9'
    TEST_MNT_POINT = TEST_MNT_BASE + '/' + TEST_HASH

    def setUp(self):
        super(BrickRemoteFsTestCase, self).setUp()
        self._mox = mox.Mox()
        self._nfsclient = remotefs.RemoteFsClient(
            'nfs', 'sudo', nfs_mount_point_base=self.TEST_MNT_BASE)
        self.addCleanup(self._mox.UnsetStubs)

    def test_get_hash_str(self):
        """_get_hash_str should calculation correct value."""

        self.assertEqual(self.TEST_HASH,
                         self._nfsclient._get_hash_str(self.TEST_EXPORT))

    def test_get_mount_point(self):
        mnt_point = self._nfsclient.get_mount_point(self.TEST_EXPORT)
        self.assertEqual(mnt_point, self.TEST_MNT_POINT)

    def test_mount_nfs_should_mount_correctly(self):
        mox = self._mox
        client = self._nfsclient

        mox.StubOutWithMock(client, '_execute')
        client._execute('mount', check_exit_code=0).AndReturn(("", ""))
        client._execute('mkdir', '-p', self.TEST_MNT_POINT,
                        check_exit_code=0).AndReturn(("", ""))
        client._execute('mount', '-t', 'nfs', '-o', 'vers=4,minorversion=1',
                        self.TEST_EXPORT,
                        self.TEST_MNT_POINT,
                        root_helper='sudo', run_as_root=True,
                        check_exit_code=0).AndReturn(("", ""))
        mox.ReplayAll()

        client.mount(self.TEST_EXPORT)

        mox.VerifyAll()

    def test_mount_nfs_with_specific_vers(self):
        opts = ['vers=2,nointr', 'nfsvers=3,lock', 'nolock,v2', 'v4.0']
        for opt in opts:
            client = remotefs.RemoteFsClient(
                'nfs', 'sudo', nfs_mount_point_base=self.TEST_MNT_BASE,
                nfs_mount_options=opt)

            client._read_mounts = mock.Mock(return_value=[])
            client._execute = mock.Mock(return_value=True)

            client.mount(self.TEST_EXPORT)
            client._execute.assert_any_call('mkdir', '-p', self.TEST_MNT_POINT,
                                            check_exit_code=0)
            client._execute.assert_any_call('mount', '-t', 'nfs', '-o',
                                            opt, self.TEST_EXPORT,
                                            self.TEST_MNT_POINT,
                                            root_helper='sudo',
                                            run_as_root=True,
                                            check_exit_code=0)

    def test_mount_nfs_with_fallback_no_vers(self):
        def execute(*args, **kwargs):
            if 'mkdir' in args:
                return True
            elif 'mount' in args:
                if 'lock,nointr,vers=4,minorversion=1' in args:
                    raise Exception()
                else:
                    return True
            else:
                self.fail(_("Unexpected call to _execute."))

        opts = 'lock,nointr'
        client = remotefs.RemoteFsClient(
            'nfs', 'sudo', nfs_mount_point_base=self.TEST_MNT_BASE,
            nfs_mount_options=opts)

        client._read_mounts = mock.Mock(return_value=[])
        client._execute = mock.Mock(wraps=execute)

        client.mount(self.TEST_EXPORT)
        client._execute.assert_any_call('mkdir', '-p', self.TEST_MNT_POINT,
                                        check_exit_code=0)
        client._execute.assert_any_call('mount', '-t', 'nfs', '-o',
                                        'lock,nointr,vers=4,minorversion=1',
                                        self.TEST_EXPORT,
                                        self.TEST_MNT_POINT,
                                        root_helper='sudo',
                                        run_as_root=True,
                                        check_exit_code=0)
        client._execute.assert_any_call('mount', '-t', 'nfs', '-o',
                                        'lock,nointr',
                                        self.TEST_EXPORT,
                                        self.TEST_MNT_POINT,
                                        root_helper='sudo',
                                        run_as_root=True,
                                        check_exit_code=0)

    def test_mount_nfs_with_fallback_all_fail(self):
        def execute(*args, **kwargs):
            if 'mkdir' in args:
                return True
            else:
                raise Exception(_("mount failed."))

        opts = 'lock,nointr'
        client = remotefs.RemoteFsClient(
            'nfs', 'sudo', nfs_mount_point_base=self.TEST_MNT_BASE,
            nfs_mount_options=opts)

        client._read_mounts = mock.Mock(return_value=[])
        client._execute = mock.Mock(wraps=execute)
        self.assertRaises(exception.BrickException, client.mount,
                          self.TEST_EXPORT)

    def test_mount_nfs_should_not_remount(self):
        mox = self._mox
        client = self._nfsclient

        line = "%s on %s type nfs (rw)\n" % (self.TEST_EXPORT,
                                             self.TEST_MNT_POINT)
        mox.StubOutWithMock(client, '_execute')
        client._execute('mount', check_exit_code=0).AndReturn((line, ""))
        mox.ReplayAll()

        client.mount(self.TEST_EXPORT)

        mox.VerifyAll()

    def test_nfs_mount_options(self):
        opts = 'test_nfs_mount_options'
        client = remotefs.RemoteFsClient(
            'nfs', 'sudo', nfs_mount_point_base=self.TEST_MNT_BASE,
            nfs_mount_options=opts)
        self.assertEqual(opts, client._mount_options)

    def test_nfs_mount_point_base(self):
        base = '/mnt/test/nfs/mount/point/base'
        client = remotefs.RemoteFsClient('nfs', 'sudo',
                                         nfs_mount_point_base=base)
        self.assertEqual(base, client._mount_base)

    def test_glusterfs_mount_point_base(self):
        base = '/mnt/test/glusterfs/mount/point/base'
        client = remotefs.RemoteFsClient('glusterfs', 'sudo',
                                         glusterfs_mount_point_base=base)
        self.assertEqual(base, client._mount_base)



#    Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from hashlib import md5
import urllib2

from lxml import etree


class HPMSAConnectionError(Exception):
    pass


class HPMSAAuthenticationError(Exception):
    pass


class HPMSARequestError(Exception):
    pass


class HPMSAClient(object):
    def __init__(self, host, login, password, protocol='http'):
        self._login = login
        self._password = password
        self._base_url = "%s://%s/api" % (protocol, host)
        self._session_key = None

    def _get_auth_token(self, xml):
        """Parse an XML authentication reply to extract the session key."""

        self._session_key = None
        obj = etree.XML(xml).find("OBJECT")
        for prop in obj.iter("PROPERTY"):
            if prop.get("name") == "response":
                self._session_key = prop.text
                break

    def login(self):
        """Authenticates the service on the device."""
        hash = md5("%s_%s" % (self._login, self._password))
        digest = hash.hexdigest()

        url = self._base_url + "/login/" + digest
        try:
            xml = urllib2.urlopen(url).read()
        except urllib2.URLError:
            raise HPMSAConnectionError()

        self._get_auth_token(xml)

        if self._session_key is None:
            raise HPMSAAuthenticationError()

    def _assert_response_ok(self, tree):
        """Parses the XML returned by the device to check the return code.

        Raises a HPMSARequestError error if the return code is not 0.
        """

        for obj in tree.iter():
            if obj.get("basetype") != "status":
                continue

            ret_code = ret_str = None
            for prop in obj.iter("PROPERTY"):
                if prop.get("name") == "return-code":
                    ret_code = prop.text
                elif prop.get("name") == "response":
                    ret_str = prop.text

            if ret_code != "0":
                raise HPMSARequestError(ret_str)
            else:
                return

        raise HPMSARequestError("No status found")

    def _build_request_url(self, path, args=None, **kargs):
        url = self._base_url + path
        if kargs:
            url += '/' + '/'.join(["%s/%s" % (k.replace('_', '-'), v)
                                   for (k, v) in kargs.items()])
        if args:
            if not isinstance(args, list):
                args = [args]
            url += '/' + '/'.join(args)

        return url

    def _request(self, path, args=None, **kargs):
        """Performs an HTTP request on the device.

        Raises a HPMSARequestError if the device returned but the status is
        not 0. The device error message will be used in the exception message.

        If the status is OK, returns the XML data for further processing.
        """

        url = self._build_request_url(path, args, **kargs)
        headers = {'dataType': 'api', 'sessionKey': self._session_key}
        req = urllib2.Request(url, headers=headers)
        try:
            xml = urllib2.urlopen(req).read()
        except urllib2.URLError:
            raise HPMSAConnectionError()

        try:
            tree = etree.XML(xml)
        except etree.LxmlError:
            raise HPMSAConnectionError()

        self._assert_response_ok(tree)
        return tree

    def logout(self):
        url = self._base_url + '/exit'
        try:
            urllib2.urlopen(url)
            return True
        except HPMSARequestError:
            return False

    def create_volume(self, vdisk, name, size):
        # NOTE: size is in this format: [0-9]+GB
        self._request("/create/volume", name, vdisk=vdisk, size=size)
        return None

    def delete_volume(self, name):
        self._request("/delete/volumes", name)

    def extend_volume(self, name, added_size):
        self._request("/expand/volume", name, size=added_size)

    def create_snapshot(self, volume_name, snap_name):
        self._request("/create/snapshots", snap_name, volumes=volume_name)

    def delete_snapshot(self, snap_name):
        self._request("/delete/snapshot", ["cleanup", snap_name])

    def vdisk_exists(self, vdisk):
        try:
            self._request("/show/vdisks", vdisk)
            return True
        except HPMSARequestError:
            return False

    def vdisk_stats(self, vdisk):
        stats = {'free_capacity_gb': 0,
                 'total_capacity_gb': 0}
        tree = self._request("/show/vdisks", vdisk)

        for obj in tree.iter():
            if obj.get("basetype") != "virtual-disks":
                continue

            for prop in obj.iter("PROPERTY"):
                # the sizes are given in number of blocks of 512 octets
                if prop.get("name") == "size-numeric":
                    stats['total_capacity_gb'] = \
                        int(prop.text) * 512 / (10 ** 9)
                elif prop.get("name") == "freespace-numeric":
                    stats['free_capacity_gb'] = \
                        int(prop.text) * 512 / (10 ** 9)

        return stats

    def _get_first_available_lun_for_host(self, host):
        luns = []
        tree = self._request("/show/host-maps", host)

        for obj in tree.iter():
            if obj.get("basetype") != "host-view-mappings":
                continue

            for prop in obj.iter("PROPERTY"):
                if prop.get("name") == "lun":
                    luns.append(int(prop.text))

        lun = 1
        while True:
            if lun not in luns:
                return lun
            lun += 1

    def map_volume(self, volume_name, wwpns):
        # NOTE(gpocentek): we assume that luns will be the same for all hosts
        lun = self._get_first_available_lun_for_host(wwpns[0])
        hosts = ",".join(wwpns)
        self._request("/map/volume", volume_name,
                      lun=str(lun), host=hosts, access="rw")
        return lun

    def unmap_volume(self, volume_name, wwpns):
        hosts = ",".join(wwpns)
        self._request("/unmap/volume", volume_name, host=hosts)

    def get_active_target_ports(self):
        ports = []
        tree = self._request("/show/ports")

        for obj in tree.iter():
            if obj.get("basetype") != "port":
                continue

            port = {}
            for prop in obj.iter("PROPERTY"):
                prop_name = prop.get("name")
                if prop_name in ["port-type", "target-id", "status"]:
                    port[prop_name] = prop.text
            if port['status'] != 'Up':
                continue
            ports.append(port)

        return ports

    def get_active_fc_target_ports(self):
        ports = []
        for port in self.get_active_target_ports():
            if port['port-type'] == "FC":
                ports.append(port['target-id'])

        return ports

    def copy_volume(self, source_name, target_name, vdisk):
        self._request("/volumecopy", target_name,
                      dest_vdisk=vdisk,
                      source_volume=source_name,
                      prompt='yes')




# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit Tests for remote procedure calls using queue
"""


import mock
import mox
from oslo.config import cfg

from cinder import context
from cinder import db
from cinder import exception
from cinder import manager
from cinder import service
from cinder import test
from cinder import wsgi


test_service_opts = [
    cfg.StrOpt("fake_manager",
               default="cinder.tests.test_service.FakeManager",
               help="Manager for testing"),
    cfg.StrOpt("test_service_listen",
               default=None,
               help="Host to bind test service to"),
    cfg.IntOpt("test_service_listen_port",
               default=0,
               help="Port number to bind test service to"), ]

CONF = cfg.CONF
CONF.register_opts(test_service_opts)


class FakeManager(manager.Manager):
    """Fake manager for tests."""
    def __init__(self, host=None,
                 db_driver=None, service_name=None):
        super(FakeManager, self).__init__(host=host,
                                          db_driver=db_driver)

    def test_method(self):
        return 'manager'


class ExtendedService(service.Service):
    def test_method(self):
        return 'service'


class ServiceManagerTestCase(test.TestCase):
    """Test cases for Services."""

    def test_message_gets_to_manager(self):
        serv = service.Service('test',
                               'test',
                               'test',
                               'cinder.tests.test_service.FakeManager')
        serv.start()
        self.assertEqual(serv.test_method(), 'manager')

    def test_override_manager_method(self):
        serv = ExtendedService('test',
                               'test',
                               'test',
                               'cinder.tests.test_service.FakeManager')
        serv.start()
        self.assertEqual(serv.test_method(), 'service')


class ServiceFlagsTestCase(test.TestCase):
    def test_service_enabled_on_create_based_on_flag(self):
        self.flags(enable_new_services=True)
        host = 'foo'
        binary = 'cinder-fake'
        app = service.Service.create(host=host, binary=binary)
        app.start()
        app.stop()
        ref = db.service_get(context.get_admin_context(), app.service_id)
        db.service_destroy(context.get_admin_context(), app.service_id)
        self.assertFalse(ref['disabled'])

    def test_service_disabled_on_create_based_on_flag(self):
        self.flags(enable_new_services=False)
        host = 'foo'
        binary = 'cinder-fake'
        app = service.Service.create(host=host, binary=binary)
        app.start()
        app.stop()
        ref = db.service_get(context.get_admin_context(), app.service_id)
        db.service_destroy(context.get_admin_context(), app.service_id)
        self.assertTrue(ref['disabled'])


class ServiceTestCase(test.TestCase):
    """Test cases for Services."""

    def setUp(self):
        super(ServiceTestCase, self).setUp()
        self.mox.StubOutWithMock(service, 'db')

    def test_create(self):
        host = 'foo'
        binary = 'cinder-fake'
        topic = 'fake'

        # NOTE(vish): Create was moved out of mox replay to make sure that
        #             the looping calls are created in StartService.
        app = service.Service.create(host=host, binary=binary, topic=topic)

        self.assertTrue(app)

    def test_report_state_newly_disconnected(self):
        host = 'foo'
        binary = 'bar'
        topic = 'test'
        service_create = {'host': host,
                          'binary': binary,
                          'topic': topic,
                          'report_count': 0,
                          'availability_zone': 'nova'}
        service_ref = {'host': host,
                       'binary': binary,
                       'topic': topic,
                       'report_count': 0,
                       'availability_zone': 'nova',
                       'id': 1}

        service.db.service_get_by_args(mox.IgnoreArg(),
                                       host,
                                       binary).AndRaise(exception.NotFound())
        service.db.service_create(mox.IgnoreArg(),
                                  service_create).AndReturn(service_ref)
        service.db.service_get(mox.IgnoreArg(),
                               mox.IgnoreArg()).AndRaise(Exception())

        self.mox.ReplayAll()
        serv = service.Service(host,
                               binary,
                               topic,
                               'cinder.tests.test_service.FakeManager')
        serv.start()
        serv.report_state()
        self.assertTrue(serv.model_disconnected)

    def test_report_state_newly_connected(self):
        host = 'foo'
        binary = 'bar'
        topic = 'test'
        service_create = {'host': host,
                          'binary': binary,
                          'topic': topic,
                          'report_count': 0,
                          'availability_zone': 'nova'}
        service_ref = {'host': host,
                       'binary': binary,
                       'topic': topic,
                       'report_count': 0,
                       'availability_zone': 'nova',
                       'id': 1}

        service.db.service_get_by_args(mox.IgnoreArg(),
                                       host,
                                       binary).AndRaise(exception.NotFound())
        service.db.service_create(mox.IgnoreArg(),
                                  service_create).AndReturn(service_ref)
        service.db.service_get(mox.IgnoreArg(),
                               service_ref['id']).AndReturn(service_ref)
        service.db.service_update(mox.IgnoreArg(), service_ref['id'],
                                  mox.ContainsKeyValue('report_count', 1))

        self.mox.ReplayAll()
        serv = service.Service(host,
                               binary,
                               topic,
                               'cinder.tests.test_service.FakeManager')
        serv.start()
        serv.model_disconnected = True
        serv.report_state()

        self.assertFalse(serv.model_disconnected)

    def test_service_with_long_report_interval(self):
        CONF.set_override('service_down_time', 10)
        CONF.set_override('report_interval', 10)
        service.Service.create(binary="test_service",
                               manager="cinder.tests.test_service.FakeManager")
        self.assertEqual(CONF.service_down_time, 25)


class TestWSGIService(test.TestCase):

    def setUp(self):
        super(TestWSGIService, self).setUp()
        self.stubs.Set(wsgi.Loader, "load_app", mox.MockAnything())

    def test_service_random_port(self):
        test_service = service.WSGIService("test_service")
        self.assertEqual(0, test_service.port)
        test_service.start()
        self.assertNotEqual(0, test_service.port)
        test_service.stop()


class OSCompatibilityTestCase(test.TestCase):
    def _test_service_launcher(self, fake_os):
        # Note(lpetrut): The cinder-volume service needs to be spawned
        # differently on Windows due to an eventlet bug. For this reason,
        # we must check the process launcher used.
        fake_process_launcher = mock.MagicMock()
        with mock.patch('os.name', fake_os):
            with mock.patch('cinder.service.process_launcher',
                            fake_process_launcher):
                launcher = service.get_launcher()
                if fake_os == 'nt':
                    self.assertEqual(type(launcher),
                                     service.Launcher)
                else:
                    self.assertEqual(launcher,
                                     fake_process_launcher())

    def test_process_launcher_on_windows(self):
        self._test_service_launcher('nt')

    def test_process_launcher_on_linux(self):
        self._test_service_launcher('posix')
